"""
Module for managing coordinates.

The coordmanager is a simple class for tracking multidimensional coordinates
and labels of ndarrays. Various methods exist for trimming, sorting, and
filtering the managed arrays based on coordinates.

# Initializing CoordinateManagers

Coordinate managers are initialized using the
[get_coord_manager](`dascore.core.coordmanager.get_coord_manager`) function. They take
a combination of coordinate dictionaries, dimensions, and attributes. They can
also be extracted from example patches.

```{python}
import dascore as dc
from rich import print

patch = dc.get_example_patch()
data = patch.data
cm = patch.coords
print(cm)
```

```{python}
import numpy as np
# Get array of coordinate values
time = cm['time']

# Filter data array
# this gets a new coord manager and a view of the trimmed data.
_sorted_time = np.sort(time)
t1 = _sorted_time[5]
t2 = _sorted_time[-5]
new_cm, new_data = cm.select(time=(t1, t2), array=data)
print(f"Old data shape: {data.shape}")
print(f"New data shape: {new_data.shape}")
print(new_cm)
```
"""

from __future__ import annotations

import warnings
from collections import defaultdict
from collections.abc import Mapping, Sequence
from itertools import zip_longest
from typing import Annotated, Any, TypeVar

import numpy as np
from pydantic import field_validator, model_validator
from rich.text import Text
from typing_extensions import Self

import dascore as dc
from dascore.constants import dascore_styles, select_values_description
from dascore.core.coords import BaseCoord, CoordSummary, get_coord
from dascore.exceptions import (
    CoordDataError,
    CoordError,
    CoordSortError,
    ParameterError,
    PatchBroadcastError,
)
from dascore.utils.attrs import separate_coord_info
from dascore.utils.docs import compose_docstring
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import (
    _apply_union_indexers,
    _matches_prefix_suffix,
    cached_method,
    iterate,
)
from dascore.utils.models import (
    ArrayLike,
    DascoreBaseModel,
    frozen_dict_serializer,
    frozen_dict_validator,
)

MaybeArray = TypeVar("MaybeArray", ArrayLike, np.ndarray, None)


def _ensure_1d_coord(coord, coord_name: str):
    """Ensure multi-dims are not used."""
    if not len(coord.shape) == 1:
        msg = (
            "Only 1 dimensional coordinates can be used for selection "
            f"{coord_name} has {len(coord.shape)} dimensions."
        )
        raise CoordError(msg)


def _indirect_coord_updates(cm, dim_name, coord_name: str, reduction, new_coords):
    """
    Applies trim to coordinates.

    Assumes other associated coordinates are trimmed.
    """
    other_coords = set(cm.dim_to_coord_map[dim_name]) - {coord_name}
    # perform indirect updates.
    for icoord in other_coords:
        dims, coord = new_coords[icoord]
        axis = cm.dim_map[icoord].index(dim_name)
        new = coord.index(reduction, axis=axis)
        new_coords[icoord] = (dims, new)


def _get_indexers_and_new_coords_dict(
    cm,
    kwargs,
    samples=False,
    relative=False,
    operation="select",
):
    """Get reductions for each dimension."""
    dim_reductions = {x: slice(None, None) for x in cm.dims}
    dimap = cm.dim_map
    new_coords = dict(cm._get_dim_array_dict(keep_coord=True))
    for coord_name, vals in kwargs.items():
        # this is not a selectable coord, just skip.
        if coord_name not in cm.coord_map or not len(cm.dim_map[coord_name]):
            continue
        coord = cm.coord_map[coord_name]
        _ensure_1d_coord(coord, coord_name)
        dim_name = dimap[coord_name][0]
        # different logic if we are using indices or values
        method = getattr(coord, operation)
        new_coord, reductions = method(vals, relative=relative, samples=samples)
        # this handles the case of out-of-bound selections.
        # These should be converted to degenerate coords.
        dim_reductions[dim_name] = reductions
        new_coords[coord_name] = (dimap[coord_name], new_coord)
        # update other coords affected by change.
        _indirect_coord_updates(cm, dim_name, coord_name, reductions, new_coords)
    indexers = tuple(dim_reductions[x] for x in cm.dims)
    return new_coords, indexers


class CoordManager(DascoreBaseModel):
    """
    Class for managing coordinates.

    Attributes
    ----------
    dims
        A tuple of dimension names.
    coord_map
        A mapping of {coord_name: Coord}
    dim_map
        A mapping of {coord_name: (dimensions, ...)}
    """

    dims: tuple[str, ...]
    coord_map: Annotated[
        FrozenDict[str, BaseCoord],
        frozen_dict_validator,
        frozen_dict_serializer,
    ]
    dim_map: Annotated[
        FrozenDict[str, tuple[str, ...]],
        frozen_dict_validator,
        frozen_dict_serializer,
    ]

    @model_validator(mode="before")
    @classmethod
    def _validate_coords(cls, values):
        """Validate the coordinates and dimensions."""
        coord_map, dim_map = values["coord_map"], values["dim_map"]
        dims = values["dims"]
        try:
            dim_shapes = {dim: coord_map[dim].shape for dim in dims}
        except KeyError:
            missing = set(dims) - set(coord_map)
            msg = f"All dimensions must have coordinates, {missing} are missing."
            raise CoordError(msg)
        # ensure non-dimensional coordinates have the same length as
        # corresponding coordinates.
        for name, coord_dims in dim_map.items():
            expected_shape = tuple(dim_shapes[x][0] for x in coord_dims)
            shape = coord_map[name].shape
            if tuple(expected_shape) == shape or not expected_shape:
                continue
            msg = (
                f"coordinate: {name} has a shape of {shape} which does not "
                f"match the dimension(s) of {coord_dims} which have a shape "
                f"of {expected_shape}"
            )
            raise CoordError(msg)
        # Ensure dimensional coordinates are associated as such
        for dim in dims:
            assert dim_map[dim] == (dim,), f"Incorrect dim association on {dim}"

        return values

    @field_validator("coord_map", "dim_map")
    @classmethod
    def _convert_to_frozen_dicts(cls, v):
        """Ensure mapping fields are immutable."""
        return FrozenDict(v)

    def __getitem__(self, item) -> np.ndarray:
        # in order to not break backward compatibility, we need to return
        # the array. Start the deprecation cycle to be consistent.
        msg = (
            "Currently coords[coord_name] returns a numpy array, but in a "
            "future dascore version it will return a BaseCoord instance. "
            "Use patch.coords.get_array(coord_name) instead."
        )
        warnings.warn(msg, UserWarning, stacklevel=2)
        return self.get_coord(item).values

    def __getattr__(self, item) -> BaseCoord:
        try:
            return super().__getattr__(item)
        except AttributeError:
            # unlike get item, get attr returns the base coordinate.
            try:
                return self.get_coord(item)
            except CoordError as e:
                raise AttributeError(e.args[0])

    def __iter__(self):
        yield from self.coord_map.items()

    def __contains__(self, key):
        return key in self.coord_map

    def update(self, **kwargs) -> Self:
        """
        Update the coordinates, return a new Coordinate Manager.

        Input values can be of the same form as initialization.
        To drop coordinates, simply pass {coord_name: None}
        """

        def _get_dim_change_drop(coord_map, dim_map):
            """
            Determine which coords must be dropped because their corresponding
            dimension size changed.
            """
            out = []
            for i, _v in coord_map.items():
                if i not in self.dims:
                    continue
                assert len(dim_map[i]) == 1, "only dealing with 1D dimensions"
                if len(coord_map[i]) == len(self.coord_map[i]):
                    continue
                for coord_name, dims in self.dim_map.items():
                    if i in dims:
                        out.append(coord_name)
            return out

        def _divide_kwargs(kwargs):
            """Divide kwargs into coords to update, drop, add, etc."""
            prefix = self.dims
            suffix = ("step", "min", "max")
            coord_updates = {
                x: kwargs.pop(x)
                for x in set(kwargs)
                if _matches_prefix_suffix(x, prefixes=prefix, suffixes=suffix)
            }
            coords_to_drop = [i for i, v in kwargs.items() if v is None]
            # convert input to coord_map/dim_map
            coords_to_add = {i: v for i, v in kwargs.items() if v is not None}
            return coord_updates, coords_to_drop, coords_to_add

        coord_updates, coord_to_drop, coord_to_add = _divide_kwargs(kwargs)
        # get coords to drop from selecting None
        coord_map, dim_map, dims = _get_coord_dim_map(coord_to_add, self.dims)
        # find coords to drop because their dimension changed.
        indirect_coord_drops = _get_dim_change_drop(coord_map, dim_map)
        # drop coords then call get_coords to handle adding new ones.
        coords, _ = self.drop_coords(*(coord_to_drop + indirect_coord_drops))
        out = coords._get_dim_array_dict(keep_coord=True)
        out.update({i: v for i, v in kwargs.items() if i not in coord_to_drop})
        # update based on keywords
        for item, value in coord_updates.items():
            coord_name, attr = item.split("_")
            new = list(out[coord_name])
            new[1] = new[1].update(**{attr: value})
            out[coord_name] = tuple(new)

        dims = tuple(x for x in dims if x not in coord_to_drop)
        return get_coord_manager(out, dims=dims)

    # we need this here to maintain backwards compatibility
    update_coords = update

    def update_from_attrs(
        self, attrs: Mapping | dc.PatchAttrs
    ) -> tuple[Self, dc.PatchAttrs]:
        """
        Update coordinates from attrs.

        This will also return a PatchAttrs which conforms to coords.

        Parameters
        ----------
        attrs
            The attribute source, either PatchAttrs instance or mapping.
        """
        coord_info, attr_info = separate_coord_info(attrs, dims=self.dims)
        out = dict(self.coord_map)
        for name in set(coord_info) & set(out):
            maybe_updates = coord_info[name]
            coord = self.coord_map[name]
            # convert values to dict to determine which should be updated.
            model_contents = coord.to_summary().model_dump(exclude_defaults=True)
            # see what has changed.
            diff = {
                i: v for i, v in maybe_updates.items() if v != model_contents.get(i)
            }
            out[name] = coord.update(**diff)
        coords = self.new(coord_map=out)
        # anything not used in coord_info should be put back.
        # for example, data_units might get put in its own coord called data.
        for unused_dim in set(coord_info) - set(coords.dims):
            for key, val in coord_info[unused_dim].items():
                attr_info[f"{unused_dim}_{key}"] = val
        attr_info["coords"] = coords.to_summary_dict()
        attr_info["dims"] = coords.dims
        attrs = dc.PatchAttrs.from_dict(attr_info)
        return coords, attrs

    def sort(
        self, *coords, array: MaybeArray = None, reverse: bool = False
    ) -> tuple[Self, MaybeArray]:
        """
        Sort coordinates.

        Parameters
        ----------
        *coords
            Specify which dimensions to sort. If None, sort along dimensional
            coordinates.
        array
            The array to sort.
        reverse
            If true, sort in descending order, else ascending.

        Raises
        ------
        CoordSortError if sorting is not possible.
        """

        def _validate_coords(coords):
            """Coordinates should be one dimensional."""
            dims = [self.dim_map[x] for x in coords]
            if any(len(x) > 1 for x in dims):
                bad_coord = [y for x, y in zip(dims, coords, strict=True) if len(y) > 1]
                msg = (
                    f"cannot simultaneously sort coordinate(s) {bad_coord}"
                    f" because they are associated with more than one dimension."
                )
                raise CoordSortError(msg)
            if len(coords) == 1:  # no need to check conflict if one dim
                return
            flat_dims = [x for y in dims for x in y]
            if len(flat_dims) != len(set(flat_dims)):
                msg = f"cannot merge {coords} because they share a dimension"
                raise CoordSortError(msg)

        def _get_dimensional_sorts(coords2sort):
            """Create the indexer for each coord."""
            new_coords = dict(self.coord_map)
            indexes = []
            for name, coord in coords2sort.items():
                dim = self.dim_map[name][0]
                new_coord, indexer = coord.sort(reverse=reverse)
                new_coords[name] = new_coord
                indexes.append(self._get_indexer(self.dims.index(dim), indexer))
                # also sort related coords.
                _sort_related(name, dim, indexer, new_coords)
            return new_coords, tuple(indexes)

        def _sort_related(name, dim, indexer, new_coords):
            """Find and apply sorting to related coords."""
            related_coords = set(self.dim_to_coord_map[dim]) - {name}
            for r_coord_name in related_coords:
                coord = new_coords[r_coord_name]
                ind = self.dim_map[r_coord_name].index(dim)
                slicer = [slice(None, None) for _ in coord.shape]
                slicer[ind] = indexer
                new_coords[r_coord_name] = coord[tuple(slicer)]

        # get coords that need sorting
        attr = "reverse_sorted" if reverse else "sorted"
        coords = coords if coords else self.dims
        _validate_coords(coords)
        cmap = self.coord_map
        assert set(coords).issubset(set(cmap))
        coords2sort = {x: cmap[x] for x in coords if not getattr(cmap[x], attr)}
        new_coords, indexers = _get_dimensional_sorts(coords2sort)
        if array is not None:
            for index in indexers:
                array = array[index]
        out = self.new(coord_map=new_coords)
        assert out.shape == self.shape
        return out, array

    def snap(
        self, *coords, array: MaybeArray = None, reverse: bool = False
    ) -> tuple[Self, MaybeArray]:
        """
        Force the specified coordinates to be monotonic and evenly sampled.

        Coordinates are first sorted, then snapped.

        Parameters
        ----------
        *coords
            Specify which dimensions to sort. If None, sort along dimensional
            coordinates.
        array
            The array to sort/snap.
        reverse
            If true, sort in descending order, else ascending.
        """
        coords = self.dims if len(coords) == 0 else coords
        cm, array = self.sort(*coords, array=array, reverse=reverse)
        # now the arrays are sorted it should be correct to snap dimensions.
        cmap = dict(cm.coord_map)
        for coord_name in coords:
            cmap[coord_name] = cmap[coord_name].snap()
        out = cm.new(coord_map=cmap)
        assert out.shape == self.shape
        return out, array

    def new(self, dims=None, coord_map=None, dim_map=None) -> Self:
        """
        Return a new coordmanager with specified attributes replaced.

        Parameters
        ----------
        dims
            A tuple of dimension strings.
        coord_map
            A mapping of {name: coord}
        dim_map
            A mapping of {coord_name:
        """
        out = self.__class__(
            dims=dims if dims is not None else self.dims,
            coord_map=coord_map if coord_map is not None else self.coord_map,
            dim_map=dim_map if dim_map is not None else self.dim_map,
        )
        return out

    def drop_coords(
        self,
        *coords: str | Sequence[str],
        array: MaybeArray = None,
    ) -> tuple[Self, MaybeArray]:
        """
        Drop one or more coordinates.

        If the coordinate is a dimension, also drop other coords that depend
        on that dimension.

        Parameters
        ----------
        *coords
            The name of the coordinate or dimension.
        """
        dim_drop_list = []
        coords_to_drop = {x for x in iterate(coords)}
        # If there are either no coords to drop or this cm doesn't have them.
        if not coords_to_drop or not (set(self.coord_map) & coords_to_drop):
            return self, array
        # figure out which coords to drop are dimensional coordinates.
        for name, dims in self.dim_map.items():
            if coords_to_drop & set(dims):
                dim_drop_list.append(name)
        dims_to_drop = set(dim_drop_list)
        # combine dims and plain coordinates that should be dropped.
        total_drops = dims_to_drop | coords_to_drop
        coord_map = {i: v for i, v in self.coord_map.items() if i not in total_drops}
        dim_map = {i: v for i, v in self.dim_map.items() if i not in total_drops}
        dims = tuple(x for x in self.dims if x not in dims_to_drop)
        index = tuple(
            slice(None, None) if x not in coords_to_drop else slice(0, 0)
            for x in self.dims
        )
        new = self.__class__(coord_map=coord_map, dim_map=dim_map, dims=dims)
        return new, _apply_union_indexers(index, array)

    def disassociate_coord(self, *coord: str) -> Self:
        """
        Disassociate some coordinates from dimensions.

        These coordinates will no longer be associated with a dimension in
        the coord manager but can still be retrieved.

        Parameters
        ----------
        coord
            The coordinate name(s) to disassociated from their dimensions.
        """
        new = {x: (None, self.coord_map[x]) for x in coord}
        return self.drop_coords(*coord)[0].update(**new)

    def drop_disassociated_coords(self) -> Self:
        """Drop all coordinates not associated with a dimension."""
        cmap = self.coord_map
        dim_map = self.dim_map
        no_dim_coords = [x for x in cmap if dim_map[x] == ()]
        return self.drop_coords(*no_dim_coords)

    def drop_private_coords(self, array=None) -> Self:
        """Drop all coordinates whose name begin with an underscore."""
        cmap = self.coord_map
        private = tuple(x for x in cmap.keys() if x.startswith("_"))
        return self.drop_coords(*private, array=array)

    def set_dims(self, **kwargs: str) -> Self:
        """
        Set dimension to non-dimensional coordinate.

        Parameters
        ----------
        kwargs
            A mapping indicating old_dim: new_dim where new_dim refers to
            the name of a non-dimensional coordinate.
        """
        dims = list(self.dims)
        coord_map = dict(self.coord_map)
        dim_map = dict(self.dim_map)
        for old_dim, new_dim in kwargs.items():
            if new_dim not in coord_map or old_dim not in dims:
                msg = (
                    f"{old_dim} is not a dimension or {new_dim} is not a "
                    f"coordinate."
                )
                raise CoordError(msg)
            # ensure coords have the same shape
            old_coord, new_coord = coord_map[old_dim], coord_map[new_dim]
            if not old_coord.shape == new_coord.shape or len(new_coord.shape) != 1:
                msg = (
                    f"New coordinate {new_coord} with a shape of {new_coord.shape} "
                    f"does not match the shape of {old_coord} ({old_coord.shape})"
                )
                raise CoordError(msg)

            # swap out old dims
            dims[dims.index(old_dim)] = new_dim
        # now loop over dim_maps and swap out old dims with new
        old_to_new = {i: v for i, v in zip(self.dims, dims, strict=True)}
        for coord_name, coord_dims in dim_map.items():
            dim_map[coord_name] = tuple(old_to_new[x] for x in coord_dims)
        return self.__class__(dims=dims, coord_map=coord_map, dim_map=dim_map)

    def _get_single_dim_kwarg_list(self, kwargs):
        """Get a list of dicts where each dict uses a dimension at most once."""
        used_coords = sorted(set(self.coord_map) & set(kwargs))
        dims = [self.dim_map[x] for x in used_coords]
        # No duplicate usage, just return.
        if len(set(dims)) == len(dims):
            return [{i: kwargs[i] for i in used_coords}]
        # We need to split kwargs up so each dimension is used no more
        # than once in each dict (element of the list).
        dim_dicts = defaultdict(list)
        for coord, dim in zip(used_coords, dims):
            dim_dicts[dim].append(coord)
        out = []
        for args in zip_longest(*dim_dicts.values()):
            out.append({x: kwargs[x] for x in args if x is not None})
        return out

    @compose_docstring(select_desc=select_values_description)
    def select(
        self, array: MaybeArray = None, relative=False, samples=False, **kwargs
    ) -> tuple[Self, MaybeArray]:
        """
        Perform selection on coordinates.

        {select_desc}

        Parameters
        ----------
        array
            An array to which the selection will be applied.
        relative
            If True, coordinate updates are relative. Does nothing if values
            passed are numpy arrays.
        samples
            If True, the query meaning is in samples.
        **kwargs
            Used to specify dimension and select arguments.

        See also [`CoordManager.order`](`dascore.core.CoordManager.order`).
        """
        if relative or samples:
            self._check_multiple_relative(kwargs)
        # Otherwise, we need to sort through kwargs and call in a loop.
        kwarg_list = self._get_single_dim_kwarg_list(kwargs)
        for kwargs in kwarg_list:
            new_coords, indexers = _get_indexers_and_new_coords_dict(
                self, kwargs, samples=samples, relative=relative, operation="select"
            )
            self = self.update(**new_coords)
            array = _apply_union_indexers(indexers, array)
        return self, array

    @compose_docstring(select_desc=select_values_description)
    def order(
        self, array: MaybeArray = None, relative=False, samples=False, **kwargs
    ) -> tuple[Self, MaybeArray]:
        """
        Perform value-based ordering on coordinates.

        Parameters
        ----------
        array
            An array to which the selection will be applied.
        relative
            If True, coordinate updates are relative. Does nothing if values
            passed are numpy arrays.
        samples
            If True, the query meaning is in samples.
        **kwargs
            Used to specify dimension and select arguments.

        See also [`CoordManager.select`](`dascore.core.CoordManager.select`).
        """
        if relative or samples:
            self._check_multiple_relative(kwargs)
        # Otherwise, we need to sort through kwargs and call in a loop.
        kwarg_list = self._get_single_dim_kwarg_list(kwargs)
        for kwargs in kwarg_list:
            new_coords, indexers = _get_indexers_and_new_coords_dict(
                self,
                kwargs,
                samples=samples,
                relative=relative,
                operation="order",
            )
            self = self.update(**new_coords)
            array = _apply_union_indexers(indexers, array)
        return self, array

    def make_broadcastable_to(
        self,
        shape: tuple[int, ...],
        array: MaybeArray,
        drop_coords: bool = False,
    ) -> tuple[Self, MaybeArray]:
        """
        Try to make coord manager broadcastable to a given shape.

        Only dimensions with Non coords can be broadcasted up.

        Parameters
        ----------
        shape
            A shape tuple (tuple of ints)
        array
            An array with the same shape as coord manager.
        drop_coords
            If True, allow dropping coordinates to broadcast coord manager
            dimensions. Otherwise, only NonCoords can change shape.
        """
        # This guarantees the shapes are compatible
        target_shape = np.broadcast_shapes(self.shape, shape)
        # Now just determine which dims need to be expanded and if they can.
        new_coords = {}
        dims = self.dims
        for ind, (current, new) in enumerate(zip(self.shape, target_shape)):
            if current >= new:
                continue
            name = dims[ind]
            coord = self.get_coord(name)
            # We can just scale up the coord
            if coord._partial or drop_coords:
                new_coords[name] = get_coord(shape=max(current, new))
            else:
                msg = f"Cannot broadcast non-empty coord {name} to shape {new}."
                raise PatchBroadcastError(msg)
        out = array if array is None else np.broadcast_to(array, target_shape)
        return self.update_coords(**new_coords), out

    def _check_multiple_relative(self, kwargs):
        """
        Relative or sample queries cannot be performed multiple times on
        the same dimension (since multiple coords can reference the same dim).
        """
        used_dims = [self.dim_map[x] for x in kwargs if x in self.coord_map]
        if len(set(used_dims)) < len(used_dims):
            msg = f"Cannot use {kwargs} for query; some coords " f"share a dimension."
            raise CoordError(msg)

    def __rich__(self) -> str:
        """Rich formatting for the coordinate manager."""
        dc_blue = dascore_styles["dc_blue"]
        header_text = Text("➤ ") + Text("Coordinates", style=dc_blue) + Text(" (")
        lens = {x: self.coord_map[x].shape[0] for x in self.dims}
        dim_texts = Text(", ").join(
            [
                Text(x, style="bold")
                + Text(": ")
                + Text(f"{lens[x]}", style="underline")
                for x in self.dims
            ]
        )
        out = [header_text, dim_texts, ")"]
        # sort coords by dims, coords
        non_dim_coords = sorted(set(self.coord_map) - set(self.dims))
        names = list(self.dims) + non_dim_coords
        for name in names:
            # skip private coords for display
            if name.startswith("_"):
                continue
            coord = self.coord_map[name]
            coord_dims = self.dim_map[name]
            if name in self.dims:
                base = Text.assemble("\n    *", Text(name, style="bold"), ": ")
            else:
                base = Text(f"\n    {name} {coord_dims}: ")
            text = Text.assemble(base, coord.__rich__())
            out.append(text)
        return Text.assemble(*out)

    def __str__(self):
        return str(self.__rich__())

    __repr__ = __str__

    def equals(self, other) -> bool:
        """Return True if other coordinates are approx equal."""
        if not isinstance(other, self.__class__):
            return False
        if not (coord_set := set(self.coord_map)) == set(other.coord_map):
            return False
        cdict_1, cdict_2 = self.coord_map, other.coord_map
        for name in coord_set:
            if not cdict_1[name].approx_equal(cdict_2[name]):
                return False
        return True

    __eq__ = equals

    @property
    def shape(self):
        """Return the shape of the dimensions."""
        out = tuple(len(self.coord_map[x]) for x in self.dims)
        # empty arrays return (0,) as their shape, so we must do the same.
        if not out:
            return (0,)
        return out

    @property
    def size(self):
        """Return the size of the patch data matrix."""
        return np.prod(self.shape)

    @property
    def ndim(self):
        """Return the number of dimensions in the coordinage manager."""
        return len(self.dims)

    def validate_data(self, data):
        """Ensure data conforms to coordinates."""
        data = np.asarray([]) if data is None else data
        if self.shape != data.shape:
            msg = (
                f"Data array has a shape of {data.shape} which doesnt match "
                f"the coordinate manager shape of {self.shape}."
            )
            raise CoordDataError(msg)
        return data

    def _get_dim_array_dict(
        self, keep_coord=False
    ) -> dict[tuple[str], ArrayLike | BaseCoord]:
        """
        Get the coord map in the form:
        {coord_name = ((dims,), array)}.

        if keep_coord, just keep the coordinate as second arg.
        """
        out = {}
        for name, coord in self.coord_map.items():
            dims = self.dim_map[name]
            out[name] = (dims, coord if keep_coord else coord.data)
        return out

    def set_units(self, **kwargs):
        """Set the units of the coordinate manager."""
        new_coords = dict(self.coord_map)
        for name, units in kwargs.items():
            new_coords[name] = new_coords[name].set_units(units)
        return self.new(coord_map=new_coords)

    def convert_units(self, **kwargs):
        """
        Convert units in coords according to kwargs. Will raise if incompatible
        Coordinates are specified.
        """
        new_coords = dict(self.coord_map)
        for name, units in kwargs.items():
            new_coords[name] = new_coords[name].convert_units(units)
        return self.new(coord_map=new_coords)

    def simplify_units(self):
        """Simplify all units in the coordinates."""
        new_coords = dict(self.coord_map)
        for name, coord in new_coords.items():
            new_coords[name] = coord.simplify_units()
        return self.new(coord_map=new_coords)

    def transpose(self, *dims: str | type(Ellipsis)) -> Self:
        """Transpose the coordinates."""

        def _get_transpose_dims(new, old):
            """Get a full tuple of dimensions for transpose, allowing for ..."""
            new_set = set(new)
            new_list, old_list = list(new), list(old)
            if len(new_set) != len(new):
                msg = "Transpose cannot use duplicate dimensions or ... more than once."
                raise ParameterError(msg)
            # iterate new list, pop out values from old list.
            for value in new_set - {...}:
                old_list.remove(value)
            # sub in remaining values for ...
            if ... in new_set:
                ind = new_list.index(...)
                new_list = new_list[:ind] + old_list + new_list[ind + 1 :]
            # ensure all values are accounted for
            if set(new_list) != set(old):
                msg = (
                    "You must specify all dimensions in a transpose operation "
                    f"or use ... You passed {new} but dimensions are {old}"
                )
                raise ParameterError(msg)
            return tuple(new_list)

        dims = _get_transpose_dims(new=dims or self.dims[::-1], old=self.dims)
        return self.new(dims=dims)

    def rename_coord(self, **kwargs) -> Self:
        """
        Rename the coordinates or dimensions.

        Parameters
        ----------
        **kwargs
            Used to rename coord {old_coord_name: new_coord_name}
        """

        def _adjust_dim_map_to_new_dim(dim_map, old_name, new_name):
            """Get the updated dim map."""
            out = dim_map
            # replace all mentions of
            for key, value in out.items():
                if old_name not in value:
                    continue
                new_value = tuple(x if x != old_name else new_name for x in value)
                out[key] = new_value

            return out

        dims, coord_map = list(self.dims), dict(self.coord_map)
        dim_map = dict(self.dim_map)
        for old_name, new_name in kwargs.items():
            if old_name == new_name:
                continue
            if old_name in dims:  # only
                dims[dims.index(old_name)] = new_name
                dim_map = _adjust_dim_map_to_new_dim(dim_map, old_name, new_name)
            coord_map[new_name] = coord_map.pop(old_name)
            dim_map[new_name] = dim_map.pop(old_name)

        out = dict(dims=dims, coord_map=coord_map, dim_map=dim_map)
        return self.__class__(**out)

    def squeeze(self, dim: Sequence[str] | None = None) -> Self:
        """
        Squeeze length one dimensions.

        Parameters
        ----------
        dim
            The dimension name, or sequence of dimension names, to drop.
            If None, drop all length 1 dims

        Raises
        ------
        CoordError if the selected dimension has a length gt 1.
        """
        to_drop = []
        if dim is None:
            dim = [x for x in self.dims if len(self.coord_map[x]) <= 1]
        bad_dim = set(iterate(dim)) - set(self.dims)
        if bad_dim:
            msg = (
                f"The following dims cannot be dropped because they don't "
                f"exist in the coordinate manager {bad_dim}"
            )
            raise CoordError(msg)
        for name in iterate(dim):
            coord = self.coord_map[name]
            if len(coord) > 1:
                msg = f"cant squeeze dim {name} because it has non-zero length"
                raise CoordError(msg)
            to_drop.append(name)
        return self.drop_coords(*to_drop)[0]

    def decimate(self, **kwargs) -> tuple[Self, tuple[slice, ...]]:
        """
        Evenly subsample along some dimension.

        Parameters
        ----------
        **kwargs
            Used to specify the dimension to decimate.

        Notes
        -----
        Removes any coordinate which depended on the decimated dimension.
        """
        assert len(kwargs) == 1
        (dim, value) = next(iter(kwargs.items()))
        assert dim in self.dims
        dim_slice = slice(None, None, int(value))
        new_array = self.coord_map[dim][dim_slice]
        new = self.update(**{dim: new_array})
        slices = tuple(
            slice(None, None) if d != dim else slice(None, None, value)
            for d in new.dims
        )
        return new, slices

    @property
    @cached_method
    def coord_shapes(self) -> dict[str, tuple[int, ...]]:
        """Get a dict of {coord_name: shape}."""
        return {i: v.shape for i, v in self.coord_map.items()}

    @property
    @cached_method
    def dim_to_coord_map(self) -> FrozenDict[str, tuple[str, ...]]:
        """Get a dimension to coordinate map."""
        out = defaultdict(list)
        for coord, dims in self.dim_map.items():
            for dim in dims:
                out[dim].append(coord)
        return FrozenDict({i: tuple(v) for i, v in out.items()})

    def get_coord_tuple_map(self) -> dict[str, tuple[tuple[str, ...], BaseCoord]]:
        """
        Return a mapping of {coord_name: (dims, coord)}.

        This is helpful because it can be passed directly to
        get_coordinate_manager.
        """
        dims = self.dim_map
        coords = self.coord_map
        return {x: (dims[x], c) for x, c in coords.items()}

    def _get_coord_dims_tuple(self):
        """Return a tuple of ((coord, dims...,), ...)."""
        dim_map = self.dim_map
        return tuple((name, *dim_map[name]) for name in self.coord_map)

    def _get_indexer(self, ind: int | None = None, value=None):
        """
        Get an indexer for the appropriate data shape.

        This is useful for generating a tuple that can be used

        ind is a list of indices to substitute in values.
        """
        out = [slice(None, None) for _ in self.shape]
        out[ind] = value
        return tuple(out)

    def keys(self):
        """Return the keys (coordinates) in the coord manager."""
        return self.coord_map.keys()

    def to_summary_dict(self) -> dict[str, CoordSummary | tuple[str, ...]]:
        """Convert the contents of the coordinate manager to a summary dict."""
        dim_map = self.dim_map
        out = {}
        for name, coord in self.coord_map.items():
            out[name] = coord.to_summary(dims=dim_map[name])
        return out

    def get_coord(self, coord_name: str) -> BaseCoord:
        """
        Retrieve a single coordinate from the coordinate manager.

        Parameters
        ----------
        coord_name
            The name of the coordinate.

        Raises
        ------
        CoordError if the coordinate is not found in the coordinate manager.
        """
        if coord_name not in self.coord_map:
            msg = (
                f"No coordinate named {coord_name} in coord manager. "
                f"Valid coordinates are {list(self.coord_map)}."
            )
            raise CoordError(msg)
        return self.coord_map[coord_name]

    def min(self, coord_name: str):
        """Return the minimum value of a coordinate."""
        return self.get_coord(coord_name).min()

    def max(self, coord_name: str):
        """Return the maximum value of a coordinate."""
        return self.get_coord(coord_name).max()

    def step(self, coord_name: str):
        """Return the coordinate step."""
        return self.get_coord(coord_name).step

    def get_array(self, coord_name: str) -> np.ndarray:
        """Return the coordinate values as a numpy array."""
        return np.asarray(self.get_coord(coord_name))

    def coord_size(self, coord_name: str) -> int:
        """Return the coordinate size."""
        return self.get_coord(coord_name).size

    def coord_range(self, coord_name: str):
        """Return a scaler value for the coordinate (e.g., number of seconds)."""
        return self.get_coord(coord_name).coord_range()


def get_coord_manager(
    coords: Mapping[str, BaseCoord | np.ndarray] | CoordManager | None = None,
    dims: tuple[str, ...] | None = None,
    attrs: dc.PatchAttrs | dict[str, Any] | None = None,
    shape=None,
) -> CoordManager:
    """
    Create a coordinate manager.

    Parameters
    ----------
    coords
        Information about coordinates. These can be a mapping of the
        form: {name, array}, {name: (dim_name, array)}, or
        {name: ((dim_names,) array). Can also be a
        [`CoordManager`](`dascore.core.CoordManager`).
    dims
        Tuple specify dimension names
    attrs
        Attributes which can be used to create coordinates.
        Cannot be used with coords argument.
        If you want to update [`CoordManager`](`dascore.core.CoordManager`)
        use [`update_from_attrs`](`dascore.core.CoordManager.update_from_attrs`).
    shape
        The data array shape which will be managed by coord manager. This
        allows non-coordinate dimensions to be initiated.

    Examples
    --------
    >>> import numpy as np
    >>> import dascore as dc
    >>> from dascore.core.coordmanager import get_coord_manager
    >>> # initialize coordinates from dict of arrays
    >>> distance = np.arange(0, 100)
    >>> time = dc.to_datetime64(np.arange(0, 100_000, 1_000))
    >>> coords = {"distance": distance, "time": time}
    >>> dims = ("time", "distance")
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # add non-dimension 1D coordinate. Note the coord dict must be a tuple
    >>> # with the first element specifying the dimensional coord the
    >>> # non-dimensional coord is attached to.
    >>> latitude = np.random.random(distance.shape)
    >>> coords['latitude'] = ('distance', latitude)
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # Add two-D non-dimensional coord.
    >>> quality = np.random.random((len(distance), len(time)))
    >>> coords['quality'] = (("distance", "time"), quality)
    >>> cm = get_coord_manager(coords=coords, dims=dims)
    >>> # Get coordinate manager from typical patch attribute dict
    >>> attrs = dc.get_example_patch().attrs
    >>> cm = get_coord_manager(attrs=attrs)
    """
    if coords is not None and attrs is not None:
        msg = (
            "Cannot use both attrs and coords in get_coord_manager. "
            "Perhaps you want CoordManager.update_from_attrs?"
        )
        raise ParameterError(msg)
    # return coords if we already have a coord manager.
    if isinstance(coords, CoordManager):
        # maybe try to rename dims.
        if dims is not None and dims != coords.dims:
            kwargs = {i: v for i, v in zip(coords.dims, dims, strict=True)}
            coords = coords.rename_coord(**kwargs)
        return coords
    # this allows a simple dict without dims to be passed and dims pulled
    # from dict keys.
    if dims is None:
        if isinstance(coords, Mapping) and all(isinstance(x, str) for x in coords):
            dims = tuple(i for i, v in coords.items() if not isinstance(v, tuple))
        elif attrs is not None and "dims" in attrs or hasattr(attrs, "dims"):
            dims = tuple(attrs["dims"].split(","))
        else:
            dims = ()
    coords = {} if coords is None else coords
    coord_map, dim_map, dims = _get_coord_dim_map(coords, dims)
    # Add missing dims to coord map so they get set as non_coords.
    if shape and (missing_dims := (set(dims) - set(coord_map))):
        for name in missing_dims:
            coord_map[name] = get_coord(shape=shape[dims.index(name)])
            dim_map[name] = (name,)
    if attrs:
        coord_updates, _ = separate_coord_info(attrs, dims)
        updateable_coords = set(coord_updates) - set(coord_map)
        for name in updateable_coords:
            coord_map[name] = get_coord(**coord_updates[name])
            dim_map[name] = (name,)
    out = CoordManager(coord_map=coord_map, dim_map=dim_map, dims=dims)
    return out


def _get_coord_dim_map(coords, dims):
    """Get coord_map, dim_map, and new dims from coord input."""

    def _get_coord(coord):
        """Get a coordinate from various inputs."""
        if hasattr(coord, "model_dump"):
            coord = coord.model_dump(exclude_defaults=True)
        if isinstance(coord, Mapping):  # input is a dict
            out = get_coord(**coord)
        else:
            out = get_coord(data=coord)
        return out

    def _coord_from_simple(name, coord):
        """Get coordinates from {coord_name: coord} where coord_name is dim name."""
        if name not in dims:
            msg = (
                "Coordinates that are not named the same as dimensions "
                "must be passed as a tuple of the form: "
                "(dimension, coord) "
            )
            raise CoordError(msg)
        out = _get_coord(coord)
        return out, (name,)

    def _maybe_coord_from_nested(name, coord, new_dims):
        """
        Get coordinates from {coord_name: (dim_name, coord)} or
        {coord_name: ((dim_names...,), coord)}.
        """
        if not len(coord) == 2:
            msg = (
                "Second input for coords must be length two of the form:"
                " (dimension, coord) or ((dimensions,...), coord)"
            )
            raise CoordError(msg)
        dim_names = tuple(i for i in iterate(coord[0]) if i)
        # # all dims must be in the input dims or a new coord.
        d1, d2 = set(dim_names), set(dims)
        if (not d1.issubset(d2)) and d1 != {name}:
            bad_dims = d1 - d2
            msg = (
                f"Coordinate specified invalid dimension(s) {bad_dims}."
                f" Valid dimensions are {dims}"
            )
            raise CoordError(msg)
        cval = coord[1]
        # pull out any relevant info from attrs.
        coord_out = _get_coord(cval)
        # check if this is added a new dimension.
        if len(dim_names) == 1 and (newdname := dim_names[0]) == name:
            if newdname not in dims:
                new_dims.append(newdname)
        expected_shape = (cval,) if isinstance(cval, int) else np.shape(cval)
        assert coord_out.shape == expected_shape
        return coord_out, dim_names

    assert not isinstance(coords, CoordManager)
    # this is a dead code path, but may be needed later, leaving for now.
    # if isinstance(coords, CoordManager):
    #     coords_dump = coords.model_dump()
    #     return dict(coords_dump["coord_map"]), dict(coords_dump["dim_map"])

    c_map, d_map, new_dims = {}, {}, []
    # iterate coords, get coordinate output.
    for name, coord in coords.items():
        if not isinstance(coord, tuple):
            c_map[name], d_map[name] = _coord_from_simple(name, coord)
        else:
            c_map[name], d_map[name] = _maybe_coord_from_nested(name, coord, new_dims)
    if new_dims:
        dims = tuple(list(dims) + new_dims)
    return c_map, d_map, dims



"""A 2D trace object."""

from __future__ import annotations

import warnings
from collections.abc import Mapping, Sequence

import numpy as np
from rich.text import Text
from typing_extensions import Self

import dascore as dc
import dascore.proc.coords
import dascore.utils.io
from dascore import transform
from dascore.compat import DataArray, array
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import CoordManager, get_coord_manager
from dascore.core.coords import BaseCoord
from dascore.utils.display import array_to_text, attrs_to_text, get_dascore_text
from dascore.utils.models import ArrayLike
from dascore.utils.patch import check_patch_attrs, check_patch_coords, get_patch_names
from dascore.utils.time import to_float
from dascore.viz import VizPatchNameSpace


class Patch:
    """
    A Class for managing data and metadata.

    See the [patch tutorial](/tutorial/patch.qmd) for examples.

    Parameters
    ----------
    data
        The array data representing fiber optic measurements.
    coords
        The coordinates, or dimensional labels for the data.
        A few types of input are permitted. If a mapping (eg dict) the value
        should conform to one of the following three forms:
        {coord_name: coord}
        {coord_name: ((dimensions,), coord)}
        {coord_name: (dimensions, coord)}
        Where coord can be a numpy array or a
        [`BaseCoord`](`dascore.core.coords.BaseCoord`) object.
        A [`CoordManager`](`dascore.core.coordmanager.CoordManager`) is also acceptable.
    dims
        A sequence of dimension strings. The first entry corresponds to the
        first axis of data, the second to the second dimension, and so on.
    attrs
        Optional attributes (non-coordinate metadata) passed as a dict or
        [PatchAttrs](`dascore.core.attrs.PatchAttrs')

    Notes
    -----
    - If coordinates and dims are not provided, they will be extracted from
    attrs, if possible.

    - If coords and attrs are provided, attrs will have priority. This means
    if there is a conflict between information contained in both, the coords
    will be recalculated.
    """

    data: ArrayLike
    coords: CoordManager
    dims: tuple[str, ...]
    attrs: PatchAttrs | Mapping

    def __init__(
        self,
        data: ArrayLike | DataArray | None = None,
        coords: Mapping[str, ArrayLike | BaseCoord] | CoordManager | None = None,
        dims: Sequence[str] | None = None,
        attrs: Mapping | PatchAttrs | None = None,
    ):
        if isinstance(data, DataArray | self.__class__):
            data, attrs, coords = data.data, data.attrs, data.coords
        if dims is None and isinstance(coords, CoordManager):
            dims = coords.dims
        # Try to generate coords from ranges in attrs
        if coords is None and attrs is not None:
            attrs = dc.PatchAttrs.from_dict(attrs)
            coords = attrs.coords_from_dims()
            dims = dims if dims is not None else attrs.dim_tuple
        # Ensure required info is here
        non_attrs = [x is None for x in [data, coords, dims]]
        if any(non_attrs) and not all(non_attrs):
            msg = "data, coords, and dims must be defined to init Patch."
            raise ValueError(msg)

        shape = None if not hasattr(data, "shape") else data.shape
        coords = get_coord_manager(coords, dims=dims, shape=shape)
        # the only case we allow attrs to include coords is if they are both
        # dicts, in which case attrs might have unit info for coords.
        if isinstance(attrs, Mapping) and attrs:
            coords, attrs = coords.update_from_attrs(attrs)
        else:
            # ensure attrs conforms to coords
            attrs = dc.PatchAttrs.from_dict(attrs).update(coords=coords)
        assert coords.dims == attrs.dim_tuple, "dim mismatch on coords and attrs"
        self._coords = coords
        self._attrs = attrs
        self._data = array(self.coords.validate_data(data))

    def __eq__(self, other):
        """Compare one Patch."""
        return dascore.proc.equals(self, other)

    def __add__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.add)

    def __sub__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.subtract)

    def __floordiv__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.floor_divide)

    def __truediv__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.divide)

    def __mul__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.multiply)

    def __pow__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.power)

    def __mod__(self, other):
        return dascore.proc.apply_ufunc(self, other, np.power)

    # Also add reverse operators

    __radd__ = __add__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rpow__ = __pow__
    __rtruediv__ = __truediv__
    __rfloordiv__ = __floordiv__
    __rmod__ = __mod__

    def __neg__(self):
        return self.update(data=-self.data)

    # Numpy Compatibility things
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Called when a numpy array is ufunc'ed against a patch.
        """
        assert method == "__call__", "only call supported."
        # pull out patch and other thing.
        arg1, arg2, *extras = inputs
        out = dascore.proc.basic.apply_ufunc(arg1, arg2, ufunc, *extras, **kwargs)
        return out

    def __array__(self, dtype=None, copy=None):
        """Used to convert Patches to arrays."""
        data = self.data
        # dascore.utils.misc.to_object_array stores patch references in numpy
        # object arrays. When that function is called it tries to make a copy
        # of the array data with dtype == object, which takes a TON of memory.
        # For now, just don't let this method convert to object dtype arrays.
        if np.issubdtype(dtype, np.dtype(object)):
            dtype = None
        out = data.astype(dtype) if dtype is not None else data
        out = out if not copy else np.copy(out)
        return out

    def __rich__(self):
        dascore_text = get_dascore_text()
        patch_text = Text("Patch ⚡", style="bold")
        header = Text.assemble(dascore_text, " ", patch_text)
        line = Text("-" * len(header))
        coords = self.coords.__rich__()
        attrs = self.attrs
        data = array_to_text(self.data, units=attrs.get("data_units"))
        attrs = attrs_to_text(self.attrs)
        out = Text("\n").join([header, line, coords, data, attrs])
        return out

        pass

    def __str__(self):
        out = self.__rich__()
        return str(out)

    __repr__ = __str__

    @property
    def dims(self) -> tuple[str, ...]:
        """Return the dimensions contained in patch."""
        return self.coords.dims

    @property
    def ndim(self) -> int:
        """Return the number of dimensions contained in patch."""
        return len(self.coords.dims)

    @property
    def coord_shapes(self) -> dict[str, tuple[int, ...]]:
        """Return a dict of {coordinate: (shape, ...)}."""
        return self.coords.coord_shapes

    @property
    def attrs(self) -> PatchAttrs:
        """Return the dimensions contained in patch."""
        return self._attrs

    @property
    def coords(self) -> CoordManager:
        """Return the dimensions contained in patch."""
        return self._coords

    @property
    def data(self) -> ArrayLike:
        """Return the data contained in patch."""
        return self._data

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the data array."""
        return self.coords.shape

    @property
    def size(self) -> tuple[int, ...]:
        """Return the size of the data array."""
        return self.coords.size

    @property
    def seconds(self) -> float:
        """Return number of seconds in the time coordinate."""
        return to_float(self.coords.coord_range("time"))

    @property
    def channel_count(self) -> int:
        """Return number of channels in the distance coordinate."""
        return self.coords.coord_size("distance")

    # --- basic patch functionality.

    update = dascore.proc.update
    # Before 0.1.0 update was called new, this is for backwards compatibility.
    new = dascore.proc.update
    equals = dascore.proc.equals
    update_attrs = dascore.proc.update_attrs
    check_coords = check_patch_coords
    check_attrs = check_patch_attrs
    get_coord = dascore.proc.get_coord
    get_array = dascore.proc.get_array
    pipe = dascore.proc.pipe
    set_dims = dascore.proc.set_dims
    squeeze = dascore.proc.coords.squeeze
    append_dims = dascore.proc.coords.append_dims
    transpose = dascore.proc.coords.transpose
    add_distance_to = dascore.proc.coords.add_distance_to
    snap_coords = dascore.proc.snap_coords
    sort_coords = dascore.proc.sort_coords
    rename_coords = dascore.proc.rename_coords
    update_coords = dascore.proc.update_coords
    drop_coords = dascore.proc.drop_coords
    drop_private_coords = dascore.proc.drop_private_coords
    coords_from_df = dascore.proc.coords_from_df
    make_broadcastable_to = dascore.proc.make_broadcastable_to
    apply_ufunc = dascore.proc.apply_ufunc
    get_patch_names = get_patch_names

    def get_patch_name(self, *args, **kwargs) -> str:
        """
        Return the name of the patch.

        See [`get_patch_names`](`dascore.utils.patch.get_patch_names`)
        for argument details.
        """
        return get_patch_names(self, *args, **kwargs).iloc[0]

    def assign_coords(self, *args, **kwargs):
        """Deprecated method for update_coords."""
        msg = "assign_coords is deprecated, use update_coords instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.update_coords(*args, **kwargs)

    set_units = dascore.proc.set_units
    convert_units = dascore.proc.convert_units
    simplify_units = dascore.proc.simplify_units

    # --- processing funcs

    select = dascore.proc.select
    order = dascore.proc.order

    def iselect(self, *args, **kwargs):
        """Deprecated  form of select."""
        msg = "patch.iselect is deprecated. Use patch.select with samples=True"
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.select(*args, samples=True, **kwargs)

    correlate = dascore.proc.correlate
    correlate_shift = dascore.proc.correlate_shift
    decimate = dascore.proc.decimate
    detrend = dascore.proc.detrend
    dropna = dascore.proc.dropna
    fillna = dascore.proc.fillna
    pass_filter = dascore.proc.pass_filter
    sobel_filter = dascore.proc.sobel_filter
    median_filter = dascore.proc.median_filter
    notch_filter = dascore.proc.notch_filter
    savgol_filter = dascore.proc.savgol_filter
    gaussian_filter = dascore.proc.gaussian_filter
    slope_filter = dascore.proc.slope_filter
    abs = dascore.proc.abs
    conj = dascore.proc.conj
    real = dascore.proc.real
    imag = dascore.proc.imag
    angle = dascore.proc.angle
    resample = dascore.proc.resample
    pad = dascore.proc.pad
    roll = dascore.proc.roll

    def iresample(self, *args, **kwargs):
        """Deprecated method."""
        msg = (
            "patch.iresample is deprecated. Please use patch.resample "
            "with samples=True"
        )
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        return self.resample(*args, samples=True, **kwargs)

    interpolate = dascore.proc.interpolate
    normalize = dascore.proc.normalize
    standardize = dascore.proc.standardize
    taper = dascore.proc.taper
    taper_range = dascore.proc.taper_range
    rolling = dascore.proc.rolling
    whiten = dascore.proc.whiten

    # --- Patch aggregations shortcuts.
    aggregate = dascore.proc.agg.aggregate
    min = dascore.proc.agg.min
    max = dascore.proc.agg.max
    mean = dascore.proc.agg.mean
    median = dascore.proc.agg.median
    std = dascore.proc.agg.std
    first = dascore.proc.agg.first
    last = dascore.proc.agg.last
    sum = dascore.proc.agg.sum

    # --- transformation functions
    differentiate = transform.differentiate
    rfft = transform.rfft
    dft = transform.dft
    idft = transform.idft
    integrate = transform.integrate
    spectrogram = transform.spectrogram
    velocity_to_strain_rate = transform.velocity_to_strain_rate
    velocity_to_strain_rate_edgeless = transform.velocity_to_strain_rate_edgeless
    dispersion_phase_shift = transform.dispersion_phase_shift
    tau_p = transform.tau_p

    # --- Method Namespaces
    # Note: these can't be cached_property (from functools) or references
    # to self stick around and keep large arrays in memory.

    @property
    def viz(self) -> VizPatchNameSpace:
        """The visualization namespace."""
        return VizPatchNameSpace(self)

    @property
    def tran(self) -> Self:
        """The transformation namespace."""
        msg = (
            "The tran namespace is deprecated. Its methods can now be "
            "accessed as normal path methods (eg patch.dft)"
        )
        warnings.warn(msg, DeprecationWarning)
        return self

    @property
    def io(self) -> dc.io.PatchIO:
        """Return a patch IO object for saving patches to various formats."""
        return dc.io.PatchIO(self)



"""Tests for DASDAE format."""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest

import dascore as dc
from dascore.compat import random_state
from dascore.io.dasdae.core import DASDAEV1
from dascore.utils.misc import register_func
from dascore.utils.time import to_datetime64

# a list of fixture names for written DASDAE files
WRITTEN_FILES = []


@pytest.fixture(scope="class")
@register_func(WRITTEN_FILES)
def written_dascore_v1_random(random_patch, tmp_path_factory):
    """Write the example patch to disk."""
    path = tmp_path_factory.mktemp("dasdae_file") / "test.hdf5"
    dc.write(random_patch, path, "dasdae", file_version="1")
    return path


@pytest.fixture(scope="class")
@register_func(WRITTEN_FILES)
def written_dascore_v1_random_indexed(written_dascore_v1_random, tmp_path_factory):
    """Copy the previous dasdae file and create an index."""
    new_path = tmp_path_factory.mktemp("dasdae_test_path") / "indexed_dasdae.h5"
    shutil.copy(written_dascore_v1_random, new_path)
    # index new path
    DASDAEV1().index(new_path)
    return new_path


@pytest.fixture(scope="class")
@register_func(WRITTEN_FILES)
def written_dascore_v1_empty(tmp_path_factory):
    """Write an empty patch to the dascore format."""
    path = tmp_path_factory.mktemp("empty_patcc") / "empty.hdf5"
    patch = dc.Patch()
    dc.write(patch, path, "DASDAE", file_version="1")
    return path


@pytest.fixture(scope="class")
@register_func(WRITTEN_FILES)
def written_dascore_correlate(tmp_path_factory, random_patch):
    """Write a correlate patch to the dascore format."""
    path = tmp_path_factory.mktemp("correlate_patcc") / "correlate.hdf5"
    padded_pa = random_patch.pad(time="correlate")
    dft_pa = padded_pa.dft("time", real=True)
    cc_pa = dft_pa.correlate(distance=[0, 1, 2], samples=True)
    dc.write(cc_pa, path, "DASDAE", file_version="1")
    return path


@pytest.fixture(params=WRITTEN_FILES, scope="class")
def dasdae_v1_file_path(request):
    """Gatherer fixture to iterate through each written dasedae format."""
    return request.getfixturevalue(request.param)


class TestWriteDASDAE:
    """Ensure the format can be written."""

    def test_file_exists(self, dasdae_v1_file_path):
        """The file should *of course* exist."""
        assert Path(dasdae_v1_file_path).exists()

    def test_append(self, written_dascore_v1_random, tmp_path_factory, random_patch):
        """Ensure files can be appended to unindexed dasdae file."""
        # make a copy of the dasdae file.
        new_path = tmp_path_factory.mktemp("dasdae_append") / "tmp.h5"
        shutil.copy(written_dascore_v1_random, new_path)
        # ensure the patch exists in the copied spool.
        df_pre = dc.spool(new_path).get_contents()
        assert len(df_pre) == 1
        # append patch to dasdae file
        new_patch = random_patch.update_attrs(time_min="1990-01-01")
        dc.write(new_patch, new_path, "DASDAE")
        # ensure the file has grown in contents
        df = dc.spool(new_path).get_contents()
        assert len(df) == len(df_pre) + 1
        assert (df["time_min"] == to_datetime64("1990-01-01")).any()

    def test_append_with_index(
        self, written_dascore_v1_random_indexed, tmp_path_factory, random_patch
    ):
        """Ensure patches can be appended to indexed dasdae file."""
        # make a copy of the dasdae file.
        new_path = tmp_path_factory.mktemp("dasdae_append") / "tmp.h5"
        shutil.copy(written_dascore_v1_random_indexed, new_path)
        # ensure the patch exists in the copied spool.
        df_pre = dc.spool(new_path).get_contents()
        assert len(df_pre) == 1
        # append patch to dasdae file
        new_patch = random_patch.update_attrs(time_min="1990-01-01")
        dc.write(new_patch, new_path, "DASDAE")
        # ensure the file has grown in contents
        df = dc.spool(new_path).get_contents()
        assert len(df) == len(df_pre) + 1
        assert (df["time_min"] == to_datetime64("1990-01-01")).any()

    def test_write_again(self, written_dascore_v1_random, random_patch):
        """Ensure a patch can be written again to file (should overwrite old)."""
        random_patch.io.write(written_dascore_v1_random, "dasdae")
        read_patch = dc.spool(written_dascore_v1_random)[0]
        assert random_patch == read_patch

    def test_write_cc_patch(self, written_dascore_correlate):
        """Ensure cross correlated patches can be writen and read."""
        sp_cc = dc.spool(written_dascore_correlate)
        assert isinstance(sp_cc[0], dc.Patch)


class TestReadDASDAE:
    """Test for reading a dasdae format."""

    def test_round_trip_random_patch(self, random_patch, tmp_path_factory):
        """Ensure the random patch can be round-tripped."""
        path = tmp_path_factory.mktemp("dasedae_round_trip") / "rt.h5"
        dc.write(random_patch, path, "DASDAE")
        out = dc.read(path)
        assert len(out) == 1
        assert out[0].equals(random_patch)

    def test_round_trip_empty_patch(self, written_dascore_v1_empty):
        """Ensure an emtpy patch can be deserialize."""
        spool = dc.read(written_dascore_v1_empty)
        assert len(spool) == 1
        spool[0].equals(dc.Patch())

    def test_datetimes(self, tmp_path_factory, random_patch):
        """Ensure the datetimes in the attrs come back as datetimes."""
        # create a patch with a custom dt attribute.
        path = tmp_path_factory.mktemp("dasdae_dt_saes") / "rt.h5"
        dt = np.datetime64("2010-09-12")
        patch = random_patch.update_attrs(custom_dt=dt)
        patch.io.write(path, "dasdae")
        patch_2 = dc.read(path)[0]
        # make sure custom tag with dt comes back from read.
        assert patch_2.attrs["custom_dt"] == dt
        # test coords are still dt64
        array = patch_2.coords.get_array("time")
        assert np.issubdtype(array.dtype, np.datetime64)
        # test attrs
        for name in ("time_min", "time_max"):
            assert isinstance(patch_2.attrs[name], np.datetime64)

    def test_read_file_no_wavegroup(self, generic_hdf5):
        """Ensure an h5 with no wavegroup returns empty patch."""
        parser = DASDAEV1()
        spool = parser.read(generic_hdf5)
        assert not len(spool)


class TestScanDASDAE:
    """Tests for scanning the dasdae format."""

    def test_scan_returns_info(self, written_dascore_v1_random, random_patch):
        """Ensure scanning returns expected values."""
        info1 = dc.scan(written_dascore_v1_random)[0].model_dump()
        info2 = random_patch.attrs.model_dump()
        common_keys = set(info1) & set(info2) - {"history"}
        for key in common_keys:
            assert info1[key] == info2[key]

    # TODO we need to re-think indexing before this can work.
    @pytest.mark.xfail
    def test_indexed_vs_unindexed(
        self,
        written_dascore_v1_random,
        written_dascore_v1_random_indexed,
    ):
        """Whether the file is indexed or not the summary should be the same."""
        df1 = dc.scan_to_df(written_dascore_v1_random)
        df2 = dc.scan_to_df(written_dascore_v1_random_indexed)
        # common fields should be equal (except path)
        common = list((set(df1) & set(df2)) - {"path"})
        assert df1[common].equals(df2[common])


class TestRoundTrips:
    """Tests for round-tripping various patches/spools."""

    formatter = DASDAEV1()

    def test_write_patch_with_lat_lon(
        self, random_patch_with_lat_lon, tmp_path_factory
    ):
        """
        DASDAE should support writing patches with non-dimensional
        coords.
        """
        new_path = tmp_path_factory.mktemp("dasdae_append") / "tmp.h5"
        shape = random_patch_with_lat_lon.shape
        dims = random_patch_with_lat_lon.dims
        # add time deltas to ensure they are also serialized/deserialized.
        dist_shape = shape[dims.index("distance")]
        time_deltas = dc.to_timedelta64(random_state.random(dist_shape))
        patch = random_patch_with_lat_lon.update_coords(
            delta_times=("distance", time_deltas),
        )
        dc.write(patch, new_path, "DASDAE")
        spool = dc.read(new_path, file_format="DASDAE")
        assert len(spool) == 1
        new_patch = spool[0]
        assert patch.equals(new_patch)

    def test_roundtrip_empty_time_patch(self, tmp_path_factory, random_patch):
        """A patch with a dimension of length 0 should roundtrip."""
        path = tmp_path_factory.mktemp("round_trip_time_degenerate") / "out.h5"
        patch = random_patch
        # get degenerate patch
        time = patch.get_coord("time")
        time_max = time.max() + 3 * time.step
        empty_patch = patch.select(time=(time_max, ...))
        empty_patch.io.write(path, "dasdae")
        spool = self.formatter.read(path)
        new_patch = spool[0]
        assert empty_patch.equals(new_patch)

    def test_roundtrip_dim_1_patch(self, tmp_path_factory, random_patch):
        """A patch with length 1 time axis should roundtrip."""
        path = tmp_path_factory.mktemp("round_trip_dim_1") / "out.h5"
        patch = dc.get_example_patch(
            "random_das",
            time_step=0.999767552,
            shape=(100, 1),
            time_min="2023-06-13T15:38:00.49953408",
        )
        patch.io.write(path, "dasdae")

        spool = self.formatter.read(path)
        new_patch = spool[0]
        assert patch.equals(new_patch)

    def test_roundtrip_datetime_coord(self, tmp_path_factory, random_patch):
        """Ensure a patch with an attached datetime coord works."""
        path = tmp_path_factory.mktemp("roundtrip_datetme_coord") / "out.h5"
        dist = random_patch.get_coord("distance")
        dt = dc.to_datetime64(np.zeros_like(dist))
        dt[0] = dc.to_datetime64("2017-09-17")
        new = random_patch.update_coords(dt=("distance", dt))
        new.io.write(path, "dasdae")
        patch = dc.spool(path, file_format="DASDAE")[0]
        assert isinstance(patch, dc.Patch)

    def test_roundtrip_nullish_datetime_coord(self, tmp_path_factory, random_patch):
        """Ensure a patch with an attached datetime coord with nulls works."""
        path = tmp_path_factory.mktemp("roundtrip_datetime_coord") / "out.h5"
        dist = random_patch.get_coord("distance")
        dt = dc.to_datetime64(np.zeros_like(dist))
        dt[~dt.astype(bool)] = np.datetime64("nat")
        dt[0] = dc.to_datetime64("2017-09-17")
        dt[-4] = dc.to_datetime64("2020-01-03")
        new = random_patch.update_coords(dt=("distance", dt))
        new.io.write(path, "dasdae")
        patch = dc.spool(path, file_format="DASDAE")[0]
        assert isinstance(patch, dc.Patch)

    # Frustratingly, it doesn't seem pytables can store NaN values using
    # create_array, even when specifying an Atom with dflt=np.nan. See
    # https://github.com/PyTables/PyTables/issues/423
    @pytest.mark.xfail(reason="Pytables issue 423")
    def test_roundtrip_len_1_non_coord(self, random_spool, tmp_path_factory):
        """Ensure we can round-trip Non-coords."""
        path = tmp_path_factory.mktemp("roundtrip_non_coord") / "out.h5"
        # create a spool that has all non coords
        spool = dc.spool([x.mean("time") for x in random_spool])
        in_patch = spool[0]
        in_patch.io.write(path, "dasdae")
        new_spool = dc.spool(path, file_format="DASDAE")
        out_patch = new_spool[0]
        assert in_patch == out_patch



"""IO module for reading Silixa's TDMS DAS data format."""

from __future__ import annotations

import dascore as dc
from dascore.constants import timeable_types
from dascore.core import Patch
from dascore.io import BinaryReader, FiberIO

from .utils import _get_data, _get_default_attrs, _get_version_str


class TDMSFormatterV4713(FiberIO):
    """Support for Silixa data format (tdms)."""

    name = "TDMS"
    version = "4713"
    preferred_extensions = ("tdms",)
    lead_in_length = 28

    def get_format(self, stream: BinaryReader, **kwargs) -> tuple[str, str] | bool:
        """
        Return a tuple of (TDMS, version) if TDMS else False.

        Parameters
        ----------
        stream
            A path to the file which may contain silixa data.
        """
        try:
            version_str = _get_version_str(stream)
            if version_str:
                return "TDMS", version_str
            else:
                return False
        except Exception:
            return False

    def scan(self, resource: BinaryReader, **kwargs) -> list[dc.PatchAttrs]:
        """Scan a tdms file, return summary information about the file's contents."""
        out = _get_default_attrs(resource)
        out["path"] = getattr(resource, "name", "")
        out["file_format"] = self.name
        out["file_version"] = self.version
        return [dc.PatchAttrs(**out)]

    def read(
        self,
        resource: BinaryReader,
        time: tuple[timeable_types, timeable_types] | None = None,
        distance: tuple[float, float] | None = None,
        **kwargs,
    ) -> dc.BaseSpool:
        """Read a silixa tdms file, return a DataArray."""
        # get all data, total amount of samples and associated attributes
        data, channel_length, attrs_full = _get_data(resource, lead_in_length=28)
        attrs = _get_default_attrs(resource, attrs_full)
        coords = dc.core.get_coord_manager(attrs.pop("coords"))
        # trim data if required
        if time is not None or distance is not None:
            coords, data = coords.select(data, time=time, distance=distance)
        patch = Patch(data=data, coords=coords, attrs=attrs)
        return dc.spool(patch)



"""Tests for Dispersion transforms."""

import numpy as np
import pytest

import dascore as dc
from dascore import get_example_patch
from dascore.exceptions import ParameterError
from dascore.transform import dispersion_phase_shift
from dascore.utils.misc import suppress_warnings


class TestDispersion:
    """Tests for the dispersion module."""

    @pytest.fixture(scope="class")
    def dispersion_patch(self, random_patch):
        """Return the random patched transformed to frequency-velocity."""
        test_vels = np.linspace(1500, 5000, 351)
        with suppress_warnings(DeprecationWarning):
            out = dispersion_phase_shift(random_patch, test_vels, approx_resolution=2.0)
        return out

    def test_dispersion(self, dispersion_patch):
        """Check consistency of test_dispersion module."""
        # assert velocity dimension
        assert "velocity" in dispersion_patch.dims
        # assert frequency dimension
        assert "frequency" in dispersion_patch.dims
        vels = dispersion_patch.coords.get_array("velocity")
        freqs = dispersion_patch.coords.get_array("frequency")
        assert np.array_equal(vels, np.linspace(1500, 5000, 351))
        # Check that the velocity output is correct
        assert freqs[1] - freqs[0] > 1.9 and freqs[1] - freqs[0] < 2.1
        # check that the approximate frequency resolution is obtained

    def test_dispersion_no_resolution(self, random_patch):
        """Ensure dispersion calc works when no resolution nor limits are provided."""
        test_vels = np.linspace(1500, 5000, 50)
        # create a smaller patch so this runs quicker.
        patch = random_patch.select(
            time=(0, 50), distance=(0, 10), relative=True, samples=True
        )
        dispersive_patch = dispersion_phase_shift(patch, test_vels)
        assert isinstance(dispersive_patch, dc.Patch)
        assert "velocity" in dispersive_patch.dims
        assert "frequency" in dispersive_patch.dims

    def test_non_monotonic_velocities(self, random_patch):
        """Ensure non-monotonic velocities raise Parameter Error."""
        msg = "must be monotonically increasing"
        velocities = np.array([10, -2, 100, 42])
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(phase_velocities=velocities)

    def test_velocity_lt_0_raises(self, random_patch):
        """Ensure velocity values < 0 raise ParameterError."""
        msg = "Velocities must be positive"
        velocities = np.array([-1, 0, 1])
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(phase_velocities=velocities)

    def test_approx_resolution_gt_0(self, random_patch):
        """Ensure velocity values < 0 raise ParameterError."""
        msg = "Frequency resolution has to be positive"
        test_vels = np.linspace(1500, 5000, 10)
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(
                phase_velocities=test_vels,
                approx_resolution=-1,
            )

    def test_freq_range_gt_0(self):
        """Ensure negative Parameter Error."""
        msg = "Minimal and maximal frequencies have to be positive"
        velocities = np.linspace(1500, 5000, 50)
        patch = get_example_patch("dispersion_event")
        with pytest.raises(ParameterError, match=msg):
            patch.dispersion_phase_shift(
                phase_velocities=velocities,
                approx_resolution=1.0,
                approx_freq=[-10, 50],
            )
        with pytest.raises(ParameterError, match=msg):
            patch.dispersion_phase_shift(
                phase_velocities=velocities,
                approx_resolution=1.0,
                approx_freq=[10, -50],
            )

    def test_freq_range_raises(self, random_patch):
        """Ensure negative Parameter Error."""
        msg = "Maximal frequency needs to be larger than minimal frequency"
        velocities = np.linspace(1500, 5000, 50)
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(
                phase_velocities=velocities, approx_resolution=1.0, approx_freq=[23, 22]
            )

    def test_freq_range_nyquist(self, random_patch):
        """Ensure negative Parameter Error."""
        msg = "Frequency range cannot exceed Nyquist"
        velocities = np.linspace(1500, 5000, 50)
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(
                phase_velocities=velocities,
                approx_resolution=1.0,
                approx_freq=[5000, 8000],
            )

    def test_freq_range_yields_empty(self, random_patch):
        """Ensure negative Parameter Error."""
        msg = "Combination of frequency resolution and range is not an array"
        velocities = np.linspace(1500, 5000, 50)
        with pytest.raises(ParameterError, match=msg):
            random_patch.dispersion_phase_shift(
                phase_velocities=velocities,
                approx_resolution=2.0,
                approx_freq=[22.0, 22.1],
            )



"""Tests for DASCore models and related functionality."""

from __future__ import annotations

import numpy as np

from dascore.utils.models import DascoreBaseModel, sensible_model_equals


class _TestModel(DascoreBaseModel):
    array: np.ndarray | None = None
    _private: int = 0
    some_str: str = "10"


class TestModelEquals:
    """Tests for seeing if models/dicts are equal."""

    def test_empty(self):
        """Empty dicts should be equal."""
        assert sensible_model_equals({}, {})

    def test_arrays_not_equal(self):
        """Ensure when arrays aren't equal models arent."""
        mod1 = _TestModel(array=np.arange(10))
        mod2 = _TestModel(array=np.arange(10) + 10)
        assert not sensible_model_equals(mod1, mod2)

    def test_private(self):
        """When private attrs arent equal the models should still be."""
        mod1 = _TestModel(_private=1)
        mod2 = _TestModel(_private=2)
        assert sensible_model_equals(mod1, mod2)

    def test_new(self):
        """Ensure a new model can b e created."""
        mod = _TestModel(some_str="test")
        new = mod.new(some_str="bob")
        assert new.some_str == "bob"
        assert new is not mod



"""Utilities functions for Neubrex IO support"""

import dascore as dc
from dascore.utils.misc import maybe_get_items


def _is_neubrex(h5fi):
    """Determine if the file is of Neubrex origin."""
    expected_keys = {"data", "depth", "stamps"}
    keys = set(h5fi.keys())
    if not expected_keys.issubset(keys):
        return False
    expected_attrs = {"DataUnitLabel", "StartDateTime", "EndDateTime"}
    data_attrs = set(h5fi["data"].attrs)
    if expected_attrs.issubset(data_attrs):
        return True


def _get_coord_manager(h5fi, snap=True):
    """Get a coordinate manager from the file."""

    def _get_time_coord(h5fi, snap):
        """Get the time coordinate."""
        # Unix stamps are in us for test files, not sure if always true.
        unix_stamps = dc.to_datetime64(h5fi["stamps_unix"][:] / 1_000_000)
        time_coord = dc.get_coord(values=unix_stamps)
        if snap:
            time_coord = time_coord.snap()
        return time_coord

    def _get_dist_coord(h5fi):
        """Get the distance (depth) coordinate."""
        depth = h5fi["depth"][:]
        return dc.get_coord(values=depth)

    coords = {
        "time": _get_time_coord(h5fi, snap=snap),
        "distance": _get_dist_coord(h5fi),
    }
    return dc.get_coord_manager(coords=coords, dims=("time", "distance"))


def _get_data_units_and_type(data_unit_label):
    """Get the units from contained string."""
    quantity = dc.get_quantity(data_unit_label.replace("-", ""))
    return quantity


def _get_attr_dict(h5fi):
    """Get a dict of neubrex attributes."""
    mapping = {
        "API": "api",
        # "DataUnitLabel": "data_unit_label",
        "FieldName": "field_name",
        "WellID": "well_id",
        "WellName": "well_name",
        "WellBoreID": "well_bore_id",
    }
    data_attrs = dict(h5fi["data"].attrs)
    out = maybe_get_items(data_attrs, mapping)
    out["data_units"] = _get_data_units_and_type(data_attrs["DataUnitLabel"])
    return out


def _maybe_trim_data(cm, data, time=None, distance=None, **kwargs):
    """Maybe trim the data."""
    if time is not None or distance is not None:
        cm, data = cm.select(time=time, distance=distance, array=data)
    return cm, data


def _get_attrs_coords_and_data(h5fi, snap=True):
    """Return the attributes, coordinates, and data array."""
    cm = _get_coord_manager(h5fi, snap)
    attrs = _get_attr_dict(h5fi)
    data = h5fi["data"]
    return attrs, cm, data



"""Tests for coordinate processing methods."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.compat import is_array
from dascore.core.coords import BaseCoord
from dascore.exceptions import (
    CoordError,
    ParameterError,
    PatchBroadcastError,
    PatchError,
)
from dascore.units import get_quantity


class TestSortCoords:
    """Test sorting patches' coordinates."""

    def test_forward_sort(self, wacky_dim_patch):
        """Test sort both dims forward."""
        # default sort should just sort both dims.
        out = wacky_dim_patch.sort_coords()
        for name, coord in out.coords.coord_map.items():
            if name not in out.dims:
                continue
            assert coord.sorted

    def test_reverse_sort(self, wacky_dim_patch):
        """Test sort both dims backward."""
        out = wacky_dim_patch.sort_coords(reverse=True)
        for name, coord in out.coords.coord_map.items():
            if name not in out.dims:
                continue
            assert coord.reverse_sorted

    def test_data_sorted_correctly(self, wacky_dim_patch):
        """Simple test to ensure data were sorted correctly."""
        patch = wacky_dim_patch
        shape = patch.shape
        dims = patch.dims
        # get array that counts cols and rows
        array_list = np.meshgrid(
            np.arange(shape[0]), np.arange(shape[1]), indexing="ij"
        )
        assert all([x.shape == patch.shape for x in array_list])
        # iterate each array and manually test sorting of data matches expected.
        for dim, array, ind in zip(dims, array_list, (1, 0)):
            coord = patch.coords.coord_map[dim]
            arg_sort = np.argsort(coord.values)
            new = patch.update(data=array).sort_coords(dim)
            data_along_slice = np.take(new.data, 0, ind)
            assert np.all(np.equal(arg_sort, data_along_slice))


class TestSnapDims:
    """Tests for snapping dimensions."""

    def test_snap_monotonic(self, wacky_dim_patch):
        """Ensure we can snap a single monotonic coordinate."""
        out = wacky_dim_patch.snap_coords("time")
        coord = out.coords.coord_map["time"]
        assert coord.sorted
        assert coord.evenly_sampled

    def test_snap_array(self, wacky_dim_patch):
        """Ensure we can snap a non monotonic coordinate."""
        out = wacky_dim_patch.snap_coords("distance")
        coord = out.coords.coord_map["distance"]
        assert coord.sorted
        assert coord.evenly_sampled

    def test_snap_dims(self, wacky_dim_patch):
        """Ensure we can snap a non monotonic coordinate."""
        out = wacky_dim_patch.snap_coords()
        for dim in out.dims:
            coord = out.coords.coord_map[dim]
            assert coord.sorted
            assert coord.evenly_sampled


class TestDropCoords:
    """Tests for dropping coordinates."""

    def test_drop_non_dim(self, random_patch_with_lat_lon):
        """Ensure non_dim coords can be dropped."""
        out = random_patch_with_lat_lon.drop_coords("latitude")
        assert "latitude" not in out.coords.coord_map

    def test_drop_dim_raises(self, random_patch):
        """Ensure a dimensional coordinate can be dropped."""
        msg = "Cannot drop dimensional coordinates"
        with pytest.raises(ParameterError, match=msg):
            random_patch.drop_coords("time")


class TestCoordsFromDf:
    """Tests for attaching coordinate(s) to a patch."""

    def get_line_func(self, df, x_col, y_col):
        """Get a function which predicts value of y based on x."""
        assert len(df) == 2
        x1, x2 = df[x_col].min(), df[x_col].max()
        y1, y2 = df[y_col].min(), df[y_col].max()
        slope = (y2 - y1) / (x2 - x1)
        intercept = y1 - x1 * slope

        def _func(x):
            return x * slope + intercept

        return _func

    @pytest.fixture()
    def coord_df(self, brady_hs_das_dts_coords, random_patch):
        """Get a coordinate dataframe that is compatible with random_patch."""
        dist_max = random_patch.coords.max("distance")
        df = (
            brady_hs_das_dts_coords.rename(columns={"Channel": "distance"})
            .sample(frac=1 / 5)
            .loc[lambda x: x["distance"] <= dist_max]
            .sort_values("distance")
            .reset_index(drop=True)
        )
        return df

    def test_interpolation_no_extrapolate(self, random_patch, coord_df):
        """Ensure interpolated values follow expected line without extrapolation."""
        # get a dataframe with only two points and interpolate.
        sub_df = coord_df.iloc[[0, -1]]
        out = random_patch.coords_from_df(sub_df, extrapolate=False)
        dist = out.coords.get_array("distance")
        for col in set(sub_df.columns[1:]) - set(out.dims):
            vals = out.coords.get_array(col)
            expected = self.get_line_func(sub_df, "distance", col)(dist)
            close = np.isclose(vals, expected)
            nan = np.isnan(vals)
            assert np.all(close | nan)

    def test_interpolation_with_extrapolate(self, random_patch, coord_df):
        """Ensure interpolated values follow expected line with extrapolation."""
        # get a dataframe with only two points and interpolate.
        sub_df = coord_df.iloc[[0, -1]]
        out = random_patch.coords_from_df(sub_df, extrapolate=True)
        dist = out.coords.get_array("distance")
        for col in set(sub_df.columns[1:]) - set(out.dims):
            vals = out.coords.get_array(col)
            expected = self.get_line_func(sub_df, "distance", col)(dist)
            assert np.allclose(vals, expected)

    def test_no_extrapolate(self, random_patch, coord_df):
        """Ensure when out of range, we use nans when extrapolate is False."""
        out = random_patch.coords_from_df(coord_df, extrapolate=False)
        coords = out.coords
        dist = coords.get_array("distance")
        df_dist = coord_df["distance"]
        # iterate each new coord and ensure out of bound values are NaN
        # but in bound are not.
        in_dist = (dist <= df_dist.max()) & (dist >= df_dist.min())
        for name in set(coord_df.columns) - set(random_patch.dims):
            vals = out.coords.get_array(name)
            assert np.all(np.isnan(vals[~in_dist]))
            assert np.all(~np.isnan(vals[in_dist]))

    def test_extrapolate(self, random_patch, coord_df):
        """Ensure we can extrapolate outside coordinate range."""
        out = random_patch.coords_from_df(coord_df, extrapolate=True)
        # All values should be filled in when extrapolating.
        for name in set(coord_df.columns) - set(random_patch.dims):
            vals = out.coords.get_array(name)
            assert np.all(~np.isnan(vals))

    def test_units(self, random_patch, coord_df):
        """Test passing in unit dictionary."""
        units = {x: "m" for x in coord_df.columns[1:]}
        out = random_patch.coords_from_df(coord_df, units=units, extrapolate=True)
        for char in "XYZ":
            coord = out.get_coord(char)
            assert coord.units == get_quantity("m")

    def test_no_dim_column_raises(self, random_patch, coord_df):
        """Ensure when no columns overlap with coords an error is raised."""
        bad_df = coord_df.drop(columns="distance")
        with pytest.raises(ParameterError, match="Exactly one column"):
            random_patch.coords_from_df(bad_df)


class TestSelect:
    """Tests for selecting data from patch."""

    def test_select_by_distance(self, random_patch):
        """Ensure distance can be used to filter patch."""
        dmin, dmax = 100, 200
        pa = random_patch.select(distance=(dmin, dmax))
        assert pa.data.shape < random_patch.data.shape
        # the attrs should have updated as well
        assert pa.attrs["distance_min"] >= 100
        assert pa.attrs["distance_max"] <= 200

    def test_select_by_absolute_time(self, random_patch):
        """Ensure the data can be sub-selected using absolute time."""
        shape = random_patch.data.shape
        t1 = random_patch.attrs["time_min"] + np.timedelta64(1, "s")
        t2 = t1 + np.timedelta64(3, "s")

        pa1 = random_patch.select(time=(None, t1))
        assert pa1.attrs["time_max"] <= t1
        assert pa1.data.shape < shape

        pa2 = random_patch.select(time=(t1, None))
        assert pa2.attrs["time_min"] >= t1
        assert pa2.data.shape < shape

        tr3 = random_patch.select(time=(t1, t2))
        assert tr3.attrs["time_min"] >= t1
        assert tr3.attrs["time_max"] <= t2
        assert tr3.data.shape < shape

    def test_select_out_of_bounds_time(self, random_patch):
        """Selecting out of coordinate range should leave patch unchanged."""
        # this equates to a timestamp of 1 (eg 1 sec after 1970)
        pa1 = random_patch.select(time=(1, None))
        assert pa1 == random_patch
        # it should also work with proper datetimes.
        t1 = random_patch.attrs["time_min"] - dc.to_timedelta64(1)
        pa2 = random_patch.select(time=(t1, None))
        assert pa2 == random_patch

    def test_select_distance_leaves_time_attr_unchanged(self, random_patch):
        """Ensure selecting on distance doesn't change time."""
        dist = random_patch.coords.get_array("distance")
        dist_max, dist_mean = np.max(dist), np.mean(dist)
        out = random_patch.select(distance=(dist_mean, dist_max - 1))
        assert out.attrs["time_max"] == out.coords.max("time")

    def test_select_emptify_array(self, random_patch):
        """If select range excludes data range patch should be emptied."""
        out = random_patch.select(distance=(-100, -10))
        assert len(out.shape) == len(random_patch.shape)
        deleted_axis = out.dims.index("distance")
        assert out.shape[deleted_axis] == 0
        assert np.size(out.data) == 0

    def test_select_relative_start_end(self, random_patch):
        """Ensure relative select works on start to end."""
        patch1 = random_patch.select(time=(1, -1), relative=True)
        t1 = random_patch.attrs.time_min + dc.to_timedelta64(1)
        t2 = random_patch.attrs.time_max - dc.to_timedelta64(1)
        patch2 = random_patch.select(time=(t1, t2))
        assert patch1 == patch2

    def test_select_relative_end_end(self, random_patch):
        """Ensure relative works for end to end."""
        patch1 = random_patch.select(time=(-3, -1), relative=True)
        t1 = random_patch.attrs.time_max - dc.to_timedelta64(1)
        t2 = random_patch.attrs.time_max - dc.to_timedelta64(3)
        patch2 = random_patch.select(time=(t1, t2))
        assert patch1 == patch2

    def test_select_relative_start_start(self, random_patch):
        """Ensure relative start ot start."""
        patch1 = random_patch.select(time=(1, 3), relative=True)
        t1 = random_patch.attrs.time_min + dc.to_timedelta64(1)
        t2 = random_patch.attrs.time_min + dc.to_timedelta64(3)
        patch2 = random_patch.select(time=(t1, t2))
        assert patch1 == patch2

    def test_select_relative_start_open(self, random_patch):
        """Ensure relative start to open end."""
        patch1 = random_patch.select(time=(1, None), relative=True)
        t1 = random_patch.attrs.time_min + dc.to_timedelta64(1)
        patch2 = random_patch.select(time=(t1, None))
        assert patch1 == patch2

    def test_select_relative_end_open(self, random_patch):
        """Ensure relative start to open end."""
        patch1 = random_patch.select(time=(-1, None), relative=True)
        t1 = random_patch.attrs.time_max - dc.to_timedelta64(1)
        patch2 = random_patch.select(time=(t1, None))
        assert patch1 == patch2

    def test_time_slice_samples(self, random_patch):
        """Ensure a simple time slice works."""
        pa1 = random_patch.select(time=(1, 5), samples=True)
        pa2 = random_patch.select(time=slice(1, 5), samples=True)
        assert pa1 == pa2

    def test_non_slice_samples(self, random_patch):
        """Ensure a non-slice doesnt change patch."""
        pa1 = random_patch.select(distance=(..., ...), samples=True)
        pa2 = random_patch.select(distance=(None, ...), samples=True)
        pa3 = random_patch.select(distance=slice(None, None), samples=True)
        pa4 = random_patch.select(distance=...)
        assert pa1 == pa2 == pa3 == pa4

    def test_iselect_deprecated(self, random_patch):
        """Ensure Patch.iselect raises deprecation error."""
        msg = "iselect is deprecated"
        with pytest.warns(DeprecationWarning, match=msg):
            _ = random_patch.iselect(time=(10, -10))

    def test_select_history_outside_bounds(self, random_patch):
        """Selecting outside the bounds should do nothing to history."""
        attrs = random_patch.attrs
        dt = dc.to_timedelta64(1)
        time = (attrs["time_min"] - dt, attrs["time_max"] + dt)
        dist = (attrs["distance_min"] - 1, attrs["distance_max"] + 1)
        new = random_patch.select(time=time, distance=dist)
        # if no select performed everything should be identical.
        assert new.equals(random_patch, only_required_attrs=False)

    def test_patch_non_coord(self, random_patch):
        """Test select for a patch with a non coord."""
        new_shape = tuple([*random_patch.shape, 10])
        patch = random_patch.append_dims("face_angle").make_broadcastable_to(new_shape)
        face_angle = patch.get_coord("face_angle")
        new = patch.select(face_angle=(face_angle.min(), face_angle.max()))
        assert new == patch


class TestOrder:
    """Tests for ordering Patches."""

    def test_simple_ordering(self, random_patch):
        """Ensure order changes to specify on patch."""
        dist = random_patch.get_array("distance")
        new_dist = dist[1:5][::-1]
        new = random_patch.order(distance=new_dist)
        assert np.all(new.get_array("distance") == new_dist)

    def test_duplicate_data(self, random_patch):
        """Duplicate the data along time dimension."""
        out = random_patch.order(time=[0, 0, 0], samples=True)
        assert isinstance(out, dc.Patch)

    def test_copy(self, random_patch):
        """Ensure copy creates a copy of the data array."""
        out = random_patch.order(time=[1, 2, 3], samples=True, copy=True)
        assert isinstance(out.data, np.ndarray)


class TestAppendDims:
    """Tests for appending dummy dimensions to data array."""

    def test_no_dims_unchanged_patch(self, random_patch):
        """Ensure no kwargs yields equal patches."""
        out = random_patch.append_dims()
        assert out == random_patch

    def test_flat_dimension(self, random_patch):
        """Ensure a flat dimension only expands dimensionality."""
        out = random_patch.append_dims(new=[1])
        assert len(out.shape) == (len(random_patch.shape) + 1)
        # New dim should show up at the end.
        assert out.dims[-1] == "new"
        coord = out.coords.get_array("new")
        assert np.all(coord == np.array([1]))
        # The flatten data should remain the same.
        assert np.allclose(out.data.flatten(), random_patch.data.flatten())

    def test_non_coordinate_dim(self, random_patch):
        """Ensure we can add non dimensional coordinates."""
        out = random_patch.append_dims(new=2)
        assert "new" in out.dims
        assert out.size == random_patch.size * 2
        assert out.shape[-1] == 2

    def test_expand_dims(self, random_patch):
        """Ensure dimensions can be expanded."""
        out = random_patch.append_dims(new=[1, 2])
        assert len(out.shape) == (len(random_patch.shape) + 1)
        # New dim should show up at the end.
        assert out.dims[-1] == "new"
        coord = out.coords.get_array("new")
        assert np.all(coord == np.array([1, 2]))

    def test_expand_multiple_dims(self, random_patch):
        """Ensure several dimensions can be expanded."""
        small_patch = random_patch.select(
            time=(1, 4),
            distance=(1, 6),
            samples=True,
        )
        out = small_patch.append_dims(new=[1, 2], old=[1, 2])
        assert len(out.shape) == (len(random_patch.shape) + 2)
        # New dim should show up at the end, in order.
        assert out.dims[-2:] == ("new", "old")

    def test_append_with_args(self, random_patch):
        """Ensure we can append with just the name of the dim."""
        out = random_patch.append_dims("new", "dim")
        assert list(out.dims) == [*list(random_patch.dims), "new", "dim"]

    def test_append_with_args_and_kwargs(self, random_patch):
        """Ensure we can use both kwargs and args."""
        out = random_patch.append_dims("new", new2=2)
        assert list(out.dims) == [*list(random_patch.dims), "new", "new2"]


class TestSqueeze:
    """Tests for squeeze."""

    @pytest.fixture(scope="class")
    def flat_patch(self):
        """Create a patch with a degenerate dimension."""
        data = np.atleast_2d(np.arange(10))
        coords = {"time": np.arange(10), "distance": np.array([1])}
        dims = ("distance", "time")
        out = dc.Patch(data=data, dims=dims, coords=coords)
        assert 1 in out.shape
        return out

    def test_remove_dimension(self, flat_patch):
        """Tests for removing degenerate dimensions."""
        out = flat_patch.squeeze("distance")
        assert "distance" not in out.dims
        assert len(out.data.shape) == 1, "data should be 1d"

    def test_tutorial_example(self, random_patch):
        """Ensure the tutorial snippet works."""
        patch = random_patch.select(distance=0, samples=True)
        squeezed = patch.squeeze()
        assert len(squeezed.dims) < len(patch.dims)

    def test_non_zero_length_raises(self, flat_patch):
        """Ensure squeezing a non-flat dim raises helpful error."""
        msg = "because it has non-zero length"
        with pytest.raises(CoordError, match=msg):
            flat_patch.squeeze(dim="time")


class TestGetCoord:
    """Tests for the get_coord convenience function."""

    def test_returns_coord(self, random_patch):
        """Simply ensure a coordinate is returned."""
        for dim in random_patch.dims:
            coord = random_patch.get_coord(dim)
            assert isinstance(coord, BaseCoord)

    def test_require_sorted(self, wacky_dim_patch):
        """Test required sorted raises if coord isn't sorted."""
        msg = "is not sorted"
        with pytest.raises(CoordError, match=msg):
            wacky_dim_patch.get_coord("distance", require_sorted=True)
        # but this should work
        coord = wacky_dim_patch.get_coord("time", require_sorted=True)
        assert isinstance(coord, BaseCoord)

    def test_non_existent_coord_raises(self, random_patch):
        """Ensure requesting non-existent coordinates raises CoordError."""
        msg = "not found in Patch"
        with pytest.raises(CoordError, match=msg):
            random_patch.get_coord("fire_house")

    def test_require_evenly_sampled(self, wacky_dim_patch):
        """Test required evenly sampled raises if coord isn't."""
        msg = "is not evenly sampled"
        with pytest.raises(CoordError, match=msg):
            wacky_dim_patch.get_coord("distance", require_evenly_sampled=True)
        with pytest.raises(CoordError, match=msg):
            wacky_dim_patch.get_coord("time", require_evenly_sampled=True)


class TestMakeBroadcastable:
    """Tests for making patches broadcastable to differnt shapes."""

    def test_broadcast_non_coords(self, random_patch):
        """Ensure non-coords of length 1 can broadcast."""
        collapsed_patch = random_patch.sum()
        shape = (2, 2)
        patch = collapsed_patch.make_broadcastable_to(shape)
        assert patch.shape == shape

    def test_raises_real_coord(self, random_patch):
        """If the dimension has values, it shouldn't be broadcastable."""
        patch = random_patch.select(time=1, distance=2, samples=True)
        # The shape is broadcastable, but the coords exist so it cant
        # broadcast.
        shape = (1, 2)
        msg = "Cannot broadcast non-empty coord"
        with pytest.raises(PatchBroadcastError, match=msg):
            patch.make_broadcastable_to(shape)

    def test_incompatible_shapes(self, random_patch):
        """Incompatible shapes should raise."""
        patch = random_patch.select(time=1, samples=True)
        shape = (12, 12)
        msg = "objects cannot be broadcast to a single shape"
        with pytest.raises(ValueError, match=msg):
            patch.make_broadcastable_to(shape)

    def test_broadcastable_to_current_shape(self, random_patch):
        """Making broadcastable to current shape should do nothing."""
        patch = random_patch
        out = patch.make_broadcastable_to(patch.shape)
        assert out == patch


class TestGetArray:
    """Tests for getting data/coordinate array."""

    def test_patch_data(self, random_patch):
        """Ensure no arguments returns patch data."""
        out = random_patch.get_array()
        assert out is random_patch.data

    def test_patch_coord_array(self, random_patch):
        """Ensure we can also get arrays from coordinates."""
        for dim in random_patch.dims:
            array = random_patch.get_array(dim)
            assert is_array(array)


class TestAddDistanceTo:
    """Tests for adding distance from a point to coords."""

    @pytest.fixture(scope="class")
    def shot_series(self):
        """Get the shot series."""
        shot = pd.Series({"x": 1000, "y": 42, "z": 15})
        return shot

    def test_coord(self, random_patch_with_xyz, shot_series):
        """Ensure coords have been added."""
        out = random_patch_with_xyz.add_distance_to(shot_series)
        assert "origin_distance" in out.coords.coord_map
        for name in shot_series.index:
            assert f"origin_{name}" in out.coords.coord_map

    def test_bad_name_raises(self, random_patch_with_xyz):
        """Ensure a PatchError is raised when index are not coords."""
        ser = pd.Series({"x": 1000, "y": 42, "q": 15})
        msg = "not patch coordinates"
        with pytest.raises(PatchError, match=msg):
            random_patch_with_xyz.add_distance_to(ser)

    def test_bad_association_raises(self, random_patch_with_xyz):
        """Ensure a PatchError is raised if coords dont share a dimension."""
        ser = pd.Series({"x": 1000, "y": 42, "time": 15})
        msg = "must be associated with the same dimension"
        with pytest.raises(PatchError, match=msg):
            random_patch_with_xyz.add_distance_to(ser)

    def test_sorting(self, random_patch_with_xyz, shot_series):
        """Ensure sorting can be done on the new patch."""
        out = random_patch_with_xyz.add_distance_to(shot_series)
        sorted_patch = out.sort_coords("origin_distance")
        coord = sorted_patch.get_array("origin_distance")
        assert np.all(np.sort(coord) == coord)



"""Transformations related to strain."""

from __future__ import annotations

import warnings

import dascore as dc
from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.transform.differentiate import differentiate
from dascore.utils.patch import patch_function


@patch_function(
    required_dims=("distance",),
    required_attrs={"data_type": "velocity"},
)
def velocity_to_strain_rate(
    patch: PatchType,
    step_multiple: int = 2,
    gauge_multiple: None | int = None,
    order: int = 2,
) -> PatchType:
    r"""
    Convert velocity DAS data to strain rate using central differences.

    When order=2 and step_multiple=2 the derivative for non-edge values
    is estimated by:

    $$
    \hat{f}(x) = \frac{f(x + (n/2)dx) - f(x - (n/2)dx)}{n dx}
    $$

    Where $dx$ is the distance step and $n$ is the step_multiple. Values for
    edges are estimate with the appropriate forward/backward stencils so that
    the shape of the output data match the input data. The equation
    becomes more complicated for higher order stencils.

    Parameters
    ----------
    patch
        A patch object containing DAS data. Note: attrs['data_type'] should be
        velocity.
    step_multiple
        The multiples of spatial sampling for the central averaging stencil.
        Must be even as odd values result in a staggered grid.
    gauge_multiple
        Deprecated name for step_multiple. Use that instead.
    order
        The order for the finite difference 1st derivative stencil (accuracy).
        It must be a multiple of 2

    Examples
    --------
    >>> from contextlib import suppress
    >>> import dascore as dc
    >>> from dascore.exceptions import MissingOptionalDependencyError
    >>> patch = dc.get_example_patch("deformation_rate_event_1")
    >>>
    >>> # Example 1
    >>> # Estimate the strain rate with a gauge length twice the distance step.
    >>> patch_strain = patch.velocity_to_strain_rate(step_multiple=2)
    >>>
    >>> # Example 2
    >>> # Estimate the strain rate with a 10th order filter. This will raise
    >>> # an exception if the package findiff is not installed.
    >>> with suppress(MissingOptionalDependencyError):
    ...     patch_strain = patch.velocity_to_strain_rate(order=10)
    >>>
    >>> # Example 3
    >>> # Estimate strain rate with a 4th order filter and gauge length 4 times
    >>> # the distance step.
    >>> with suppress(MissingOptionalDependencyError):
    ...     patch_strain = patch.velocity_to_strain_rate(step_multiple=4, order=4)

    Notes
    -----
    This is primarily used with Terra15 data and simply uses
    [patch.differentiate](`dascore.transform.differentiate.differentiate`)
    under the hood to calculate spatial derivatives.

    The output gauge length is equal to the step_multiple multuplied by the
    spacing along the distance coordinate, although the concept of
    gauge_length is more complex with higher oder filters. See
    @yang2022filtering for more info.

    See the [`velocity_to_strain_rate` note](docs/notes/velocity_to_strain_rate.qmd)
    for more details on step_multiple and order effects.

    The [edgeless](`dascore.Patch.velocity_to_strain_rate_edgeless`) version
    of this function removes potential edge effects and supports even and odd
    `step_multiple` values.
    """
    if gauge_multiple is not None:
        msg = "gauge_multiple will be removed in the future. Use step_multiple."
        warnings.warn(msg, DeprecationWarning)
        step_multiple = gauge_multiple * 2

    if step_multiple % 2 != 0:
        msg = (
            "Step_multiple must be even. Use velocity_to_strain_rate_edgeless "
            "if odd step multiples are required."
        )
        raise ParameterError(msg)

    coord = patch.get_coord("distance", require_evenly_sampled=True)
    step = coord.step
    patch = differentiate.func(
        patch, dim="distance", order=order, step=step_multiple // 2
    )
    new_attrs = patch.attrs.update(
        data_type="strain_rate", gauge_length=step * step_multiple
    )
    return patch.update(attrs=new_attrs)


@patch_function(
    required_dims=("distance",),
    required_attrs={"data_type": "velocity"},
)
def velocity_to_strain_rate_edgeless(
    patch: PatchType,
    step_multiple: int = 1,
) -> PatchType:
    r"""
    Estimate strain-rate using central differences.

    For odd step_multiple values this function estimates strain by taking a
    staggered central difference according to:

    $$
    \hat{f} = \frac{f(x + n * dx/2) - f(x - n * dx/2)}{dx}
    $$

    Where $dx$ is the spatial sampling and $n$ is the step_multiple. As a result
    the strain-rate between existing samples is estimated when $n$ is odd. Edges
    (points where full central differences are not possible) are discarded in
    the output.

    Parameters
    ----------
    patch
        A patch object containing DAS data. Note: attrs['data_type'] should be
        velocity.
    step_multiple
        The number of spatial sampling steps to use in the central averaging.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch("deformation_rate_event_1")
    >>>
    >>> # Example 1
    >>> # Estimate strain rate with a gauge length equal to distance step.
    >>> patch_strain = patch.velocity_to_strain_rate_edgeless(step_multiple=1)
    >>>
    >>> # Example 2
    >>> # Estimate strain rate with a gauge length 5 times the distance step.
    >>> patch_strain = patch.velocity_to_strain_rate_edgeless(step_multiple=5)

    Notes
    -----
    See [velocity_to_strain_rate](`dascore.Patch.velocity_to_strain_rate`)
    for a similar function which does not change the shape of the patch.

    The resulting gauge length is equal to the step_multiple multiplied by
    the sampling along the distance dimension.

    See the
    [`velocity_to_strain_rate` note](docs/notes/velocity_to_strain_rate.qmd)
    for more details on step_multiple and order effects.
    """
    coord = patch.get_coord("distance", require_evenly_sampled=True)
    distance_step = coord.step
    gauge_length = step_multiple * distance_step

    data_1 = patch.select(distance=(step_multiple, None), samples=True).data
    data_2 = patch.select(distance=(None, -step_multiple), samples=True).data
    strain_rate = (data_1 - data_2) / gauge_length

    # Need to get distance values between current ones.
    dists = patch.get_array("distance")
    new_dist = (dists[step_multiple:] + dists[:-step_multiple]) / 2
    new_coords = patch.coords.update(distance=new_dist)

    # Handle unit conversions.
    new_data_units = None
    data_units = dc.get_quantity(patch.attrs.data_units)
    dist_units = dc.get_quantity(patch.get_coord("distance").units)
    if data_units and dist_units:
        new_data_units = data_units / dist_units

    new_attrs = patch.attrs.update(
        data_type="strain_rate",
        gauge_length=distance_step * step_multiple,
        data_units=new_data_units,
    )

    return dc.Patch(data=strain_rate, coords=new_coords, attrs=new_attrs)



"""Spectral whitening."""

from __future__ import annotations

import numpy as np
from scipy.ndimage import uniform_filter1d

from dascore.constants import PatchType
from dascore.exceptions import ParameterError
from dascore.utils.patch import patch_function
from dascore.utils.transformatter import FourierTransformatter


def _get_dim_freq_range_from_kwargs(patch, kwargs):
    """Get the dimension and frequency range."""
    dim_set = set(patch.dims)
    # Handles the default case when no kwargs passed.
    if not kwargs:
        expected = {"time", "ft_time"} & dim_set
        if not expected:
            msg = "No dim name provided in kwargs and patch has no time dimension."
            raise ParameterError(msg)
        dim = "time"
        freq_range = None
    # A single kwarg was passed.
    elif len(kwargs) == 1:
        dim, freq_range = next(iter(kwargs.items()))
        fft_dim = FourierTransformatter().rename_dims(dim)[0]
        if dim not in dim_set and fft_dim not in dim_set:
            msg = f"passed dim of {dim} to whiten but it is not in patch dimensions."
            raise ParameterError(msg)
    else:  # Something when wrong.
        msg = (
            "Whiten kwargs must specify a single patch dimension. "
            "Double check the supported input parameters as these may have "
            f"changed. You passed {kwargs}."
        )
        raise ParameterError(msg)

    return dim, freq_range


def _get_amp_envelope(fft_patch, axis, window_len, water_level):
    """Get a smoothed amplitude envelope."""
    amp = np.abs(fft_patch.data)
    # Uniform filter is *much* faster than convolve
    uni = uniform_filter1d(amp, window_len, axis=axis, mode="wrap")
    if water_level is not None:
        # Enforce water level to avoid instability in dividing small numbers
        uni[uni < water_level * uni.max()] = water_level * uni.max()
    smoothed_amp = amp / uni
    return smoothed_amp


def _check_smooth(fft_coord, smooth_size, water_level):
    """Check the smooth size."""
    if water_level is not None:
        if not isinstance(water_level, float) or water_level < 0 or water_level > 1:
            msg = "water_level must be a float between 0 and 1."
            raise ParameterError(msg)
    if smooth_size <= 0:
        msg = "Frequency smoothing size must be positive"
        raise ParameterError(msg)
    if smooth_size >= fft_coord.max():
        msg = "Frequency smoothing size is larger than Nyquist"
        raise ParameterError(msg)


def _check_freq_range(fft_coord, freq_range):
    """Check the frequency range."""
    # Note: the freq_range can be a len 2 or 4 sequence
    frange = np.asarray(freq_range)
    diffs = frange[1:] - frange[:-1]
    min_size = fft_coord.step * 2

    if np.any(diffs < min_size):
        msg = "Frequency range is too narrow"
        raise ParameterError(msg)


@patch_function()
def whiten(
    patch: PatchType,
    smooth_size: None | float = None,
    water_level: None | float = None,
    **kwargs,
) -> PatchType:
    """
    Spectral whitening of a signal.

    The whitened signal is returned in the same domain (eq frequency or
    time domain) as the input signal. See also the
    [Whiten Processing Section](`docs/tutorial/processing.qmd`#whiten).

    Parameters
    ----------
    patch
        The patch to transform.
    smooth_size
        Size in transformed domain units (eg Hz) or samples of moving average
        window, used to compute the spectrum before whitening.
        If None, don't smooth signal which results in a uniform amplitude.
        units.
    water_level
        If used, float between 0 and 1 to stabilize frequencies with near
        zero amplitude. Does nothing if smooth_size is None.
        Values between 0.01 and 0.05 usually work well.
    **kwargs
        Used to specify the dimension range in transformed units (e.g, Hz)
        of the smoothing. Can either be a sequence of two values or
        four values to specify a taper range. Simply uses
        [Patch.taper_range](`dascore.Patch.taper_range`) under the hood.
        If kwargs are provided, try to smooth `time` or `ft_time` coords.

    Notes
    -----
    1) The FFT result is divided by the smoothed spectrum before inverting
       back to time-domain signal. The phase is not changed.

    2) Amplitude is NOT preserved

    Example
    -------
    >>> import dascore as dc
    >>>
    >>> patch = dc.get_example_patch()
    >>>
    >>> # Whiten along time dimension
    >>> white_patch = patch.whiten(time=None)
    >>>
    >>> # Band limited whitening
    >>> white_patch = patch.whiten(time=(20, 40))
    >>>
    >>> # Band limited with taper ends
    >>> white_patch = patch.whiten(time=(10, 20, 40, 60))
    >>>
    >>> # Whitening along distance with amplitude smoothing (0.1/m))
    >>> white_patch = patch.whiten(smooth_size=0.1, distance=None)
    """
    dim, freq_range = _get_dim_freq_range_from_kwargs(patch, kwargs)
    fft_dim = FourierTransformatter().rename_dims(dim)[0]
    # Get frequency domain patch
    fft_patch = patch.dft(dim, real=np.isrealobj(patch.data))
    input_patch_fft = fft_patch is patch  # if input patch had fft
    fft_coord = fft_patch.get_coord(fft_dim)
    # Get amplitude spectra if smoothing, otherwise use ones.
    if smooth_size is None:
        amp = np.ones_like(fft_patch.data)
    else:
        _check_smooth(fft_coord, smooth_size, water_level)
        axis = fft_patch.dims.index(fft_dim)
        window_len = fft_coord.get_sample_count(smooth_size, enforce_lt_coord=True)
        amp = _get_amp_envelope(fft_patch, axis, window_len, water_level)
    # Init new output from new amplitudes and old phases.
    out = fft_patch.new(data=amp * np.exp(1j * np.angle(fft_patch.data)))
    # Apply band-limited taper to remove some frequencies.
    if freq_range:
        _check_freq_range(fft_coord, freq_range)
        out = out.taper_range.func(out, **{fft_dim: freq_range})
    # Convert back to time domain if input was in time-domain.
    if not input_patch_fft:
        out = out.idft()
    return out



"""
IO module for reading Febus data.
"""

from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.constants import opt_timeable_types
from dascore.io import FiberIO
from dascore.utils.hdf5 import H5Reader
from dascore.utils.models import UTF8Str

from .utils import (
    _get_febus_version_str,
    _read_febus,
    _yield_attrs_coords,
)


class FebusPatchAttrs(dc.PatchAttrs):
    """
    Patch attrs for febus.

    Attributes
    ----------
    source
        The source designation
    zone
        The zone designations
    """

    gauge_length: float = np.nan
    gauge_length_units: str = "m"
    pulse_width: float = np.nan
    pulse_width_units: str = "m"

    group: str = ""
    source: str = ""
    zone: str = ""

    folog_a1_software_version: UTF8Str = ""


class Febus2(FiberIO):
    """Support for Febus V 2.

    This should cover all versions 2.* of the format (maybe).
    """

    name = "febus"
    preferred_extensions = ("hdf5", "h5")
    version = "2"

    def get_format(self, resource: H5Reader, **kwargs) -> tuple[str, str] | bool:
        """
        Return True if file contains febus version 8 data else False.

        Parameters
        ----------
        resource
            A path to the file which may contain terra15 data.
        """
        version_str = _get_febus_version_str(resource)
        if version_str:
            return self.name, version_str

    def scan(self, resource: H5Reader, **kwargs) -> list[dc.PatchAttrs]:
        """Scan a febus file, return summary information about the file's contents."""
        out = []
        file_version = _get_febus_version_str(resource)
        extras = {
            "path": resource.filename,
            "file_format": self.name,
            "file_version": str(file_version),
        }
        for attr, cm, _ in _yield_attrs_coords(resource):
            attr["coords"] = cm.to_summary_dict()
            attr.update(dict(extras))
            out.append(FebusPatchAttrs(**attr))
        return out

    def read(
        self,
        resource: H5Reader,
        time: tuple[opt_timeable_types, opt_timeable_types] | None = None,
        distance: tuple[float | None, float | None] | None = None,
        **kwargs,
    ) -> dc.BaseSpool:
        """Read a febus spool of patches."""
        patches = _read_febus(
            resource, time=time, distance=distance, attr_cls=FebusPatchAttrs
        )
        return dc.spool(patches)


class Febus1(Febus2):
    """Support for Febus V 1.

    This is here to support legacy febus (eg pubdas Valencia)
    """

    version = "1"



"""Tests for basic patch functions."""

from __future__ import annotations

import operator

import numpy as np
import pandas as pd
import pytest
from scipy.fft import next_fast_len

import dascore as dc
from dascore import get_example_patch
from dascore.exceptions import ParameterError, PatchBroadcastError, UnitError
from dascore.proc.basic import apply_operator
from dascore.units import furlongs, get_quantity, m, s
from dascore.utils.misc import _merge_tuples

OP_NAMES = ("add", "sub", "pow", "truediv", "floordiv", "mul", "mod")
TEST_OPS = tuple(getattr(operator, x) for x in OP_NAMES)


@pytest.fixture(scope="session")
def random_complex_patch(random_patch):
    """Swap out data for complex data."""
    shape = random_patch.shape
    rand = np.random.random
    new_data = rand(shape) + rand(shape) * 1j
    pa = random_patch.new(data=new_data)
    return pa


class TestAbs:
    """Test absolute values."""

    def test_no_negatives(self, random_patch):
        """Simply ensure the data has no negatives."""
        # add
        data = np.array(random_patch.data)
        data[:, 0] = -2
        new = random_patch.new(data=data)
        out = new.abs()
        assert np.all(out.data >= 0)

    def test_all_real(self, random_complex_patch):
        """Ensure all values are real after abs on complex data."""
        out = random_complex_patch.abs()
        assert np.issubdtype(out.data.dtype, np.float64)
        assert np.allclose(np.imag(out.data), 0)


class TestReal:
    """Test for getting real data values."""

    def test_real_from_complex(self, random_complex_patch):
        """Simply ensure data are real-valued."""
        out1 = random_complex_patch.real()
        assert np.all(out1.data == np.real(out1.data))

    def test_real_does_nothing_on_float(self, random_patch):
        """Read shouldn't do anything for float patches."""
        pa1 = random_patch.real()
        assert np.allclose(pa1.data, random_patch.data)


class TestImag:
    """Test for getting imaginary part of data values."""

    def test_imag_from_real(self, random_patch):
        """Real data should have close to 0 imaginary part."""
        out1 = random_patch.imag()
        assert np.allclose(out1.data, 0)

    def test_imag_from_complex(self, random_complex_patch):
        """Real data should have close to 0 imaginary part."""
        out1 = random_complex_patch.imag()
        assert np.allclose(out1.data, np.imag(random_complex_patch.data))


class TestsAngle:
    """Test for getting phase angle."""

    def test_angle_from_real(self, random_patch):
        """Ensure phase angles from real value are close to 0/360."""
        out = random_patch.angle()
        assert np.allclose(out.data, 0)

    def test_angle_from_complex(self, random_complex_patch):
        """Simple phase angle test."""
        out = random_complex_patch.angle()
        assert np.allclose(out.data, np.angle(random_complex_patch.data))


class TestTranspose:
    """Tests for transposing patch."""

    def test_transpose_no_args(self, random_patch):
        """Ensure transposing rotates dimensions."""
        pa = random_patch.transpose()
        assert pa.dims != random_patch.dims
        assert pa.dims == random_patch.dims[::-1]


class TestNormalize:
    """Tests for normalization."""

    def test_bad_norm_raises(self, random_patch):
        """Ensure an unsupported norm raises."""
        with pytest.raises(ValueError):
            random_patch.normalize("time", norm="bob_norm")

    def test_l2(self, random_patch):
        """Ensure after operation norms are 1."""
        dims = random_patch.dims
        # test along distance axis
        dist_norm = random_patch.normalize("distance", norm="l2")
        axis = dims.index("distance")
        norm = np.linalg.norm(dist_norm.data, axis=axis)
        assert np.allclose(norm, 1)
        # tests along time axis
        time_norm = random_patch.normalize("time", norm="l2")
        axis = dims.index("time")
        norm = np.linalg.norm(time_norm.data, axis=axis)
        assert np.allclose(norm, 1)

    def test_l1(self, random_patch):
        """Ensure after operation norms are 1."""
        dims = random_patch.dims
        # test along distance axis
        dist_norm = random_patch.normalize("distance", norm="l1")
        axis = dims.index("distance")
        norm = np.abs(np.sum(dist_norm.data, axis=axis))
        assert np.allclose(norm, 1)
        # tests along time axis
        time_norm = random_patch.normalize("time", norm="l1")
        axis = dims.index("time")
        norm = np.abs(np.sum(time_norm.data, axis=axis))
        assert np.allclose(norm, 1)

    def test_max(self, random_patch):
        """Ensure after operation norms are 1."""
        dims = random_patch.dims
        # test along distance axis
        dist_norm = random_patch.normalize("distance", norm="l1")
        axis = dims.index("distance")
        norm = np.abs(np.sum(dist_norm.data, axis=axis))
        assert np.allclose(norm, 1)
        # tests along time axis
        time_norm = random_patch.normalize("time", norm="l1")
        axis = dims.index("time")
        norm = np.abs(np.sum(time_norm.data, axis=axis))
        assert np.allclose(norm, 1)

    def test_bit(self):
        """Ensure after operation each sample is -1, 1, or 0."""
        patch = get_example_patch("dispersion_event")
        bit_norm = patch.normalize("time", norm="bit")
        assert np.all(np.unique(bit_norm.data) == np.array([-1.0, 0, 1.0]))

    def test_zero_channels(self, random_patch):
        """Ensure after operation each zero row or vector remains so."""
        zeroed_data = np.copy(random_patch.data)
        zeroed_data[0, :] = 0.0
        zeroed_data[:, 0] = 0.0
        zeroed_patch = random_patch.new(data=zeroed_data)
        for norm_type in ["l1", "l2", "max", "bit"]:
            norm = zeroed_patch.normalize("time", norm=norm_type)
            assert np.all(norm.data[0, :] == 0.0)
            assert np.all(norm.data[:, 0] == 0.0)
        for norm_type in ["l1", "l2", "max", "bit"]:
            norm = zeroed_patch.normalize("distance", norm=norm_type)
            assert np.all(norm.data[0, :] == 0.0)
            assert np.all(norm.data[:, 0] == 0.0)


class TestStandarize:
    """Tests for standardization."""

    def test_base_case(self, random_patch):
        """Ensure runs with default parameters."""
        # test along distance axis
        out = random_patch.standardize("distance")
        assert not np.any(pd.isnull(out.data))
        # test along time axis
        out = random_patch.standardize("time")
        assert not np.any(pd.isnull(out.data))

    def test_std(self, random_patch):
        """Ensure after operation standard deviations are 1."""
        dims = random_patch.dims
        # test along distance axis
        axis = dims.index("distance")
        out = random_patch.standardize("distance")
        assert np.allclose(
            np.round(np.std(out.data, axis=axis, keepdims=True), decimals=1), 1.0
        )
        # test along time axis
        out = random_patch.standardize("time")
        axis = dims.index("time")
        assert np.allclose(
            np.round(np.std(out.data, axis=axis, keepdims=True), decimals=1), 1.0
        )

    def test_mean(self, random_patch):
        """Ensure after operation means are 0."""
        dims = random_patch.dims
        # test along distance axis
        out = random_patch.standardize("distance")
        axis = dims.index("distance")
        assert np.allclose(
            np.round(np.mean(out.data, axis=axis, keepdims=True), decimals=1), 0.0
        )
        # test along time axis
        out = random_patch.standardize("time")
        axis = dims.index("time")
        assert np.allclose(
            np.round(np.mean(out.data, axis=axis, keepdims=True), decimals=1), 0.0
        )


class TestApplyUfunc:
    """Tests for applying various ufunc-type operators."""

    def test_scalar(self, random_patch):
        """Test for a single scalar."""
        new = apply_operator(random_patch, 10, np.multiply)
        assert np.allclose(new.data, random_patch.data * 10)

    def test_array_like(self, random_patch):
        """Ensure array-like operations work."""
        ones = np.ones(random_patch.shape)
        new = apply_operator(random_patch, ones, np.add)
        assert np.allclose(new.data, ones + random_patch.data)

    def test_incompatible_coords(self, random_patch):
        """Ensure un-alignable coords returns degenerate patch."""
        time = random_patch.get_coord("time")
        new_time = time.max() + time.step
        new = random_patch.update_attrs(time_min=new_time)
        out = apply_operator(new, random_patch, np.multiply)
        assert 0 in set(out.shape)

    def test_quantity_scalar(self, random_patch):
        """Ensure operators work with quantities."""
        patch = random_patch.set_units("m/s")
        other = 10 * m / s
        # first try multiply
        new = apply_operator(patch, other, np.multiply)
        new_units = get_quantity("m/s") * get_quantity("m/s")
        assert get_quantity(new.attrs.data_units) == new_units
        assert isinstance(new.data, np.ndarray)
        # try add
        new = apply_operator(patch, other, np.add)
        new_units = get_quantity("m/s")
        assert get_quantity(new.attrs.data_units) == new_units
        assert isinstance(new.data, np.ndarray)
        # and divide
        new = apply_operator(patch, other, np.divide)
        new_units = get_quantity("m/s") / get_quantity("m/s")
        assert new.attrs.data_units is None or new.attrs.data_units == new_units
        assert isinstance(new.data, np.ndarray)

    def test_unit(self, random_patch):
        """Units should only affect the unit attr."""
        patch = random_patch.set_units("m/s")
        other = m / s
        # first try multiply
        new = apply_operator(patch, other, np.multiply)
        new_units = get_quantity("m/s") * get_quantity("m/s")
        assert get_quantity(new.attrs.data_units) == new_units
        assert np.allclose(new.data, random_patch.data)
        # try add
        new = apply_operator(patch, other, np.add)
        new_units = get_quantity("m/s")
        assert get_quantity(new.attrs.data_units) == new_units
        assert np.allclose(new.data, random_patch.data + 1)
        # and divide
        new = apply_operator(patch, other, np.divide)
        new_units = get_quantity("m/s") / get_quantity("m/s")
        assert new.attrs.data_units is None or new.attrs.data_units == new_units
        assert np.allclose(new.data, random_patch.data)

    def test_patch_with_units(self, random_patch):
        """Ensure when patch units are set they are applied as well."""
        # test add
        pa1 = random_patch.set_units("m/s")
        out1 = apply_operator(pa1, pa1, np.add)
        assert get_quantity(out1.attrs.data_units) == get_quantity("m/s")
        # test multiply
        out2 = apply_operator(pa1, pa1, np.multiply)
        assert get_quantity(out2.attrs.data_units) == get_quantity("m**2/s**2")

    def test_array_with_units(self, random_patch):
        """Ensure an array with units works for multiplication."""
        patch1 = random_patch
        ones = np.ones(patch1.shape) * furlongs
        out1 = patch1 * ones
        assert get_quantity(out1.attrs.data_units) == get_quantity(furlongs)
        # test division with units
        patch2 = random_patch.set_units("m")
        out2 = patch2 / ones
        expected = get_quantity("m/furlongs")
        assert get_quantity(out2.attrs.data_units) == expected

    def test_incompatible_units(self, random_patch):
        """Ensure incompatible units raise."""
        pa1 = random_patch.set_units("m/s")
        other = 10 * get_quantity("m")
        with pytest.raises(UnitError):
            apply_operator(pa1, other, np.add)

    def test_patches_non_coords_len_1(self, random_patch):
        """Ensure patches with non-coords also work."""
        mean_patch = random_patch.mean("distance")
        out = mean_patch / mean_patch
        assert np.allclose(out.data, 1)

    def test_patches_non_coords_different_len(self, random_patch):
        """Ensure patches with non-coords of different lengths work."""
        patch_1 = random_patch.mean("distance")
        dist_ind = patch_1.dims.index("distance")
        old_shape = list(patch_1.shape)
        old_shape[dist_ind] = old_shape[dist_ind] + 2
        patch_2 = patch_1.make_broadcastable_to(tuple(old_shape))
        out = patch_1 / patch_2
        assert np.allclose(out.data, 1)
        assert out.shape == patch_2.shape

    def test_non_dim_coords(self, random_dft_patch):
        """Ensure ufuncs can still be applied to coords with non dim coords."""
        out = random_dft_patch * random_dft_patch
        assert set(out.coords.coord_map) == set(random_dft_patch.coords.coord_map)


class TestPatchBroadcasting:
    """Tests for patches broadcasting to allow operations on each other."""

    def test_broadcast_sub_patch(self, random_patch):
        """Ensure a patch with a subset of dimensions broadcasts."""
        sub_patch = random_patch.min("time")
        out = random_patch - sub_patch
        assert isinstance(out, dc.Patch)
        axis = random_patch.dims.index("time")
        to_sub = np.min(random_patch.data, axis=axis, keepdims=True)
        # Ensure the time aggregation worked.
        assert np.allclose(sub_patch.data, to_sub)
        # Then test that the resulting data is as expected.
        expected = random_patch.data - to_sub
        assert np.allclose(expected, out.data)
        # The reverse should also be true (after accounting for negative sign)
        out2 = sub_patch - random_patch
        assert out == -out2

    def test_broadcast_single_shared_dim(self, random_patch):
        """Ensure two 2d patches can get broad-casted when one dim is the same."""
        # get two smallish patches to test multidimensional broadcasting.
        patch1 = random_patch.select(time=(1, 10), distance=(1, 10), samples=True)
        patch2 = patch1.transpose("distance", "time").rename_coords(distance="length")
        expected_dims = _merge_tuples(patch1.dims, patch2.dims)
        out = patch1 * patch2
        assert out.dims == expected_dims
        assert out.shape == (9, 9, 9)

    def test_patches_different_coords_same_shape(self, random_patch):
        """Ensure the intersection of coordinates in output when coords differ."""
        distance = random_patch.get_array("distance")
        mid = distance[len(distance) // 2]
        shifted_patch = random_patch.update_coords(distance_min=mid)
        out = random_patch + shifted_patch
        # Ensure distance is equal to intersecting values
        dist_out = out.get_array("distance")
        overlap_dist = np.intersect1d(
            random_patch.get_array("distance"), out.get_array("distance")
        )
        assert np.all(dist_out == overlap_dist)

    def test_patch_broadcast_array(self, random_patch):
        """Ensure a patch is broadcastable with an array up to the same ndims."""
        should_work = (np.array(1), np.ones(1), np.ones((1, 1)))
        for ar in should_work:
            out = random_patch * ar
            assert isinstance(out, dc.Patch)

    def test_patch_broadcast_array_more_dims_raises(self, random_patch):
        """Ensure a patch cannot broadcast to an array which has more dims."""
        ar = np.ones((1, 1, 1))
        with pytest.raises(PatchBroadcastError, match="Cannot broadcast"):
            _ = random_patch * ar

    def test_broadcast_up_array(self, random_patch):
        """Ensure a patch with empty coords can broadcast up."""
        # This patch still has empty dims of len 1
        patch = random_patch.mean()
        ar = np.ones(random_patch.shape)
        for out in [patch * ar, ar * patch]:
            assert isinstance(out, dc.Patch)
            assert np.allclose(out.data, patch.data)

    def test_broadcast_with_array(self, random_patch):
        """Ensure a patch can broadcast up and results are correct."""
        agg = random_patch.min(None)
        out1 = random_patch - agg
        out2 = random_patch.data - agg
        assert np.allclose(out1.data, out2.data)

    @pytest.mark.parametrize("test_op", TEST_OPS)
    def test_broadcast_collapsed_patch(self, random_patch, test_op):
        """Ensure a collapsed patch can still broadcast."""
        collapsed_patch = random_patch.min(None)
        scalar = 10
        # Arrays should raise since we don't know the name of the
        # dims that would be expanded.
        mat1 = np.array([1, 2, 3])
        mat2 = np.arange(4).reshape(2, 2) + 1  # + 1 to avoid divide by 0
        # A collapsed patch should broadcast to all these things.
        for val in (scalar, mat1, mat2, random_patch):
            p1 = test_op(collapsed_patch, val)
            p2 = test_op(val, collapsed_patch)
            assert isinstance(p1, dc.Patch)
            assert isinstance(p2, dc.Patch)


class TestDropNa:
    """Tests for dropping nullish values in a patch."""

    @pytest.fixture(scope="session")
    def patch_with_null(self):
        """Return a patch which has nullish values."""
        return dc.get_example_patch("patch_with_null")

    @pytest.fixture(scope="class")
    def patch_3d_with_null(self, range_patch_3d):
        """Return a patch which has nullish values."""
        data = np.array(range_patch_3d.data).astype(np.float64)
        nans = [(1, 1, 1), (0, 9, 9)]
        for nan_ind in nans:
            data[nan_ind] = np.nan
        patch = range_patch_3d.update(data=data)
        return patch

    def test_drop_time_any(self, patch_with_null):
        """Ensure we can drop NaN along time axis."""
        patch = patch_with_null.dropna("time")
        # we know that each label has at least 1 nan so size should be 0
        coord = patch.get_coord("time")
        assert len(coord) == 0

    def test_drop_time_all(self, patch_with_null):
        """Ensure we can drop NaN along time axis."""
        patch = patch_with_null.dropna("time", how="all")
        # we should have dropped the first time label
        before_coord = patch_with_null.get_coord("time")
        after_coord = patch.get_coord("time")
        assert len(before_coord) == len(after_coord) + 1
        # also should have no columns with all NaN
        axis = patch_with_null.dims.index("time")
        assert axis == 1
        expected = np.all(pd.isnull(patch.data), axis=0)
        assert not np.any(expected)

    def test_drop_distance_all(self, patch_with_null):
        """Ensure we can drop NaN along distance axis."""
        patch = patch_with_null.dropna("distance", how="all")
        # we should have dropped the first time label
        before_coord = patch_with_null.get_coord("distance")
        after_coord = patch.get_coord("distance")
        assert len(before_coord) == len(after_coord) + 1
        # also should have no columns with all NaN
        axis = patch_with_null.dims.index("distance")
        assert axis == 0
        expected = np.all(pd.isnull(patch.data), axis=1)
        assert not np.any(expected)

    def test_3d(self, patch_3d_with_null):
        """Ensure 3D patches can have NaNs dropped."""
        patch = patch_3d_with_null
        # first test all; it shouldn't change anything
        out = patch.dropna("time", how="all")
        assert out.shape == patch.shape
        # then test any; it should drop 2 labels
        axis = out.dims.index("time")
        out = patch.dropna("time", how="any")
        assert out.shape[axis] == patch.shape[axis] - 2


class TestFillNa:
    """Tests for replacing nullish values in a patch."""

    @pytest.fixture(scope="session")
    def patch_with_null(self):
        """Return a patch which has nullish values."""
        return dc.get_example_patch("patch_with_null")

    @pytest.fixture(scope="class")
    def patch_3d_with_null(self, range_patch_3d):
        """Return a patch which has nullish values."""
        data = np.array(range_patch_3d.data).astype(np.float64)
        nans = [(1, 1, 1), (0, 9, 9)]
        for nan_ind in nans:
            data[nan_ind] = np.nan
        patch = range_patch_3d.update(data=data)
        return patch

    def test_fillna(self, patch_with_null):
        """Ensure we can fillna and keep the other values the same."""
        patch = patch_with_null.fillna(0)

        assert all(patch.data[pd.isnull(patch_with_null.data)] == 0)
        assert all(
            patch.data[~pd.isnull(patch_with_null.data)]
            == patch_with_null.data[~pd.isnull(patch_with_null.data)]
        )

    def test_3d(self, patch_3d_with_null):
        """Ensure 3D patches can fillna and keep the other values the same."""
        patch = patch_3d_with_null.fillna(0)

        assert all(patch.data[pd.isnull(patch_3d_with_null.data)] == 0)
        assert all(
            patch.data[~pd.isnull(patch_3d_with_null.data)]
            == patch_3d_with_null.data[~pd.isnull(patch_3d_with_null.data)]
        )


class TestPad:
    """Tests for the padding functionality in a patch."""

    def test_pad_time_dimension_samples_true(self, random_patch, samples=True):
        """Test padding the time dimension with zeros before and after."""
        padded_patch = random_patch.pad(time=(2, 3), samples=samples)
        # Check if the padding is applied correctly
        original_shape = random_patch.shape
        new_shape = padded_patch.shape
        time_axis = random_patch.dims.index("time")
        assert new_shape[time_axis] == original_shape[time_axis] + 5
        # Ensure that padded values are zeros
        assert np.all(padded_patch.select(time=(None, 2), samples=samples).data == 0)
        assert np.all(padded_patch.select(time=(-3, None), samples=samples).data == 0)

    def test_pad_distance_dimension(self, random_patch):
        """Test padding the distance with same number of zeros on both sides."""
        padded_patch = random_patch.pad(distance=7)
        original_shape = random_patch.shape
        new_shape = padded_patch.shape
        distance_axis = random_patch.dims.index("distance")
        ch_spacing = random_patch.attrs["distance_step"]
        assert (
            new_shape[distance_axis] == original_shape[distance_axis] + 14 * ch_spacing
        )
        # Ensure that padded values are zeros
        assert np.all(padded_patch.select(distance=(None, 7), samples=True).data == 0)
        assert np.all(padded_patch.select(distance=(-7, None), samples=True).data == 0)

    def test_pad_distance_dimension_expand_coords(
        self, random_patch, expand_coords=True
    ):
        """Test padding the distance with same number of zeros on both sides."""
        padded_patch = random_patch.pad(distance=4, expand_coords=expand_coords)
        original_shape = random_patch.shape
        new_shape = padded_patch.shape
        distance_axis = random_patch.dims.index("distance")
        ch_spacing = random_patch.attrs["distance_step"]
        dist_max = random_patch.attrs["distance_max"]
        assert (
            new_shape[distance_axis] == original_shape[distance_axis] + 8 * ch_spacing
        )
        # Ensure that padded values are zeros
        assert np.all(padded_patch.select(distance=(None, -1)).data == 0)
        assert np.all(padded_patch.select(distance=(dist_max + 1, None)).data == 0)

    def test_pad_multiple_dimensions_samples_true(self, random_patch, samples=True):
        """Test padding multiple dimensions with different pad values."""
        padded_patch = random_patch.pad(
            time=(6, 7), distance=(1, 4), constant_values=np.pi, samples=samples
        )
        # Check dimensions individually
        time_axis = random_patch.dims.index("time")
        distance_axis = random_patch.dims.index("distance")
        assert padded_patch.shape[time_axis] == random_patch.shape[time_axis] + 13
        assert (
            padded_patch.shape[distance_axis] == random_patch.shape[distance_axis] + 5
        )
        # Check padding values
        assert np.allclose(
            padded_patch.select(time=(None, 6), samples=samples).data, np.pi, atol=1e-6
        )
        assert np.allclose(
            padded_patch.select(time=(-7, None), samples=samples).data, np.pi, atol=1e-6
        )
        assert np.allclose(
            padded_patch.select(distance=(None, 1), samples=samples).data,
            np.pi,
            atol=1e-6,
        )
        assert np.allclose(
            padded_patch.select(distance=(-4, None), samples=samples).data,
            np.pi,
            atol=1e-6,
        )

    @pytest.mark.parametrize("length", [10, 13, 200, 293])
    def test_fft_pad(self, random_patch, length):
        """Tests for padding to next fast length."""
        # First trim patch to match length
        patch_in = random_patch.select(time=(0, length), samples=True)
        axis = patch_in.dims.index("time")
        length_old = patch_in.shape[axis]
        next_fast = next_fast_len(length_old)
        out = patch_in.pad(time="fft")
        assert out.dims == patch_in.dims, "dims should not have changed."
        assert out.shape[axis] == next_fast

    def test_correlate_pad(self, random_patch):
        """Ensure the fft correlate pad returns sensible values."""
        axis = random_patch.dims.index("time")
        length_old = random_patch.shape[axis]
        next_fast = next_fast_len(length_old * 2 - 1)
        out = random_patch.pad(time="correlate")
        assert out.shape[axis] == next_fast

    def test_pad_no_expand(self, random_patch):
        """Ensure we can pad without expanding dimensions."""
        out = random_patch.pad(time=10, samples=True, expand_coords=False)
        coord_array = out.get_array("time")
        # Check the first and last nans
        assert np.all(pd.isnull(coord_array[:10]))
        assert np.all(pd.isnull(coord_array[-10:]))
        # Check the middle values are equal to old values.
        old_array = random_patch.get_array("time")
        # The datatype should not have changed
        assert coord_array.dtype == old_array.dtype
        assert np.all(old_array == coord_array[10:-10])

    def test_pad_no_expand_int_coord(self, random_patch):
        """Ensure we can pad an integer coordinate."""
        # Ensure distances in an in coordinate.
        coord = random_patch.get_coord("distance")
        new_vals = np.arange(len(coord), dtype=np.int64)
        patch = random_patch.update_coords(distance=new_vals)
        # Apply padding, ensure NaN values appear.
        padded = patch.pad(distance=4, expand_coords=False)
        coord_array = padded.get_array("distance")
        assert np.all(pd.isnull(coord_array[:4]))
        assert np.all(pd.isnull(coord_array[-4:]))

    def test_error_on_sequence_constant_values(self, random_patch):
        """Test that providing a sequence for constant_values raises a TypeError."""
        with pytest.raises(ParameterError):
            random_patch.pad(time=(0, 5), constant_values=(0, 0))


class TestConj:
    """Tests for complex conjugate."""

    def test_imaginary_part_reversed(self, random_dft_patch):
        """Ensure the imaginary part of the array is reversed."""
        imag1 = np.imag(random_dft_patch.data)
        conj = random_dft_patch.conj()
        imag2 = np.imag(conj.data)
        assert np.allclose(imag1, -imag2)


class TestRoll:
    """Test cases for patch roll method."""

    def test_time_roll(self, random_patch):
        """Test basic sample roll."""
        rand_patcht = random_patch.transpose("time", ...)
        rolled_patch = rand_patcht.roll(time=5, samples=True)
        assert rand_patcht.shape == rolled_patch.shape
        assert np.all(rand_patcht.data[0] == rolled_patch.data[5])

    def test_dist_roll(self, random_patch):
        """Test roll when samples=False."""
        rolled_patch = random_patch.roll(distance=30, samples=False)
        coord = random_patch.get_coord("distance")
        value = coord.get_sample_count(30)
        assert random_patch.shape == rolled_patch.shape
        assert np.all(random_patch.data[0] == rolled_patch.data[value])

    def test_coord_update_roll(self, random_patch):
        """Test roll when update_coord=True."""
        patcht = random_patch.transpose("time", ...)
        rolled_patch = patcht.roll(time=5, samples=True, update_coord=True)
        assert (
            patcht.coords.get_array("time")[0]
            == rolled_patch.coords.get_array("time")[5]
        )

    def test_dist_coord_roll(self, random_patch):
        """Test roll when samples=False and coords will be updated."""
        rolled_patch = random_patch.roll(distance=30, samples=False, update_coord=True)
        coord = random_patch.get_coord("distance")
        value = coord.get_sample_count(30)
        assert (
            random_patch.coords.get_array("distance")[0]
            == rolled_patch.coords.get_array("distance")[value]
        )
